/* Complete arithmetic in genus 2
* - Romain Cosset

This file is a collaboration with Chritophe Arene

This file contains functions to compute complete addition laws with theta functions of level 4 over genus 2 curves.


TODO rewrite this without using Gaudry numbering
*/

//import "Arithmetic.m" : init_theta_point, init_theta_null_point;


function GroupToGaudry(th)
  Ab:=th`numbering;
  k:=Parent(th`coord[[0,0,0,0]]);

  t:=[k!0 : x in [1..16]];

  t[1]:=th`coord[Ab![0,0,0,0]];
  t[2]:=th`coord[Ab![0,0,1,1]];
  t[3]:=th`coord[Ab![0,0,1,0]];
  t[4]:=th`coord[Ab![0,0,0,1]];

  t[5]:=th`coord[Ab![1,0,0,0]];
  t[6]:=th`coord[Ab![1,0,0,1]];
  t[7]:=th`coord[Ab![0,1,0,0]];
  t[8]:=th`coord[Ab![1,1,0,0]];
  t[9]:=th`coord[Ab![0,1,1,0]];
  t[10]:=th`coord[Ab![1,1,1,1]];

  t[11]:=th`coord[Ab![0,1,0,1]];
  t[12]:=th`coord[Ab![0,1,1,1]];
  t[13]:=th`coord[Ab![1,0,1,0]];
  t[14]:=th`coord[Ab![1,1,1,0]];
  t[15]:=th`coord[Ab![1,0,1,1]];
  t[16]:=th`coord[Ab![1,1,0,1]];

  return t;
end function;




/*
Let D be a divisor in Mumford coordinates on the hyperelliptic curve with Rosenhain invariants Ros such that Ros[2]-Ros[1]=1.
!! Let thc be the theta constants as an AnalyticThetaNullPoint !!
 */
function MumfordToTheta(D,thc,Ros)

  if D[3] eq 0 then
    points:=[];
  elif D[3] eq 1 then
    x:=-Coefficient(D[1],0);
    y:=D[2];
    points:=[[x,y]];
  elif D[3] eq 2 then
    points:={*[x[1],Evaluate(D[2],x[1])]^^x[2] : x in RootsInSplittingField(D[1]) *};
    points:=[ x : x in points];
  end if;

  th := MumfordToLevel4ThetaPoint(2, Ros, Parent(Ros[1])!1, thc, points);
   // returns the corresponding Theta functions

  return GroupToGaudry(th);
end function;



/*
Let thc be the theta constants numbered as in Gaudry
Let X,Y be theta points in Gaudry numbering

return the 16 Baily's additions law of X and Y as a list of list Z
Z[k] is th k-th law (wich corresponds to the projective factor
        theta_k(2(x-y),Om)=0
Z[k][l] is the l-th coordinates of X+Y given by the k-th law
 */
function BailyAdd_fromTheta(X,Y,thc)

  Z:=[[Parent(X[1])!0 : x in [1..16]] : x in [1..16]];

  Z[1][1]:=1/(thc[1]^2)*(Y[1]^2*X[1]^2+Y[5]^2*X[5]^2+Y[7]^2*X[7]^2+Y[8]^2*X[8]^2+Y[3]^2*X[3]^2+Y[13]^2*X[13]^2+Y[9]^2*X[9]^2+Y[14]^2*X[14]^2+Y[4]^2*X[4]^2+Y[6]^2*X[6]^2+Y[11]^2*X[11]^2+Y[16]^2*X[16]^2+Y[2]^2*X[2]^2+Y[15]^2*X[15]^2+Y[12]^2*X[12]^2+Y[10]^2*X[10]^2);
  Z[1][2]:=1/(thc[1]*thc[2])*(2*Y[1]*Y[2]*X[2]*X[1]+2*Y[15]*Y[5]*X[5]*X[15]+2*Y[12]*Y[7]*X[7]*X[12]+2*Y[10]*Y[8]*X[8]*X[10]+2*Y[3]*Y[4]*X[4]*X[3]+2*Y[6]*Y[13]*X[13]*X[6]+2*Y[9]*Y[11]*X[11]*X[9]+2*Y[16]*Y[14]*X[14]*X[16]);
  Z[1][3]:=1/(thc[1]*thc[3])*(2*Y[1]*Y[3]*X[3]*X[1]+2*Y[13]*Y[5]*X[5]*X[13]+2*Y[9]*Y[7]*X[7]*X[9]+2*Y[14]*Y[8]*X[8]*X[14]+2*Y[2]*Y[4]*X[4]*X[2]+2*Y[15]*Y[6]*X[6]*X[15]+2*Y[12]*Y[11]*X[11]*X[12]+2*Y[10]*Y[16]*X[16]*X[10]);
  Z[1][4]:=1/(thc[1]*thc[4])*(2*Y[4]*Y[1]*X[1]*X[4]+2*Y[6]*Y[5]*X[5]*X[6]+2*Y[11]*Y[7]*X[7]*X[11]+2*Y[16]*Y[8]*X[8]*X[16]+2*Y[3]*Y[2]*X[2]*X[3]+2*Y[15]*Y[13]*X[13]*X[15]+2*Y[12]*Y[9]*X[9]*X[12]+2*Y[10]*Y[14]*X[14]*X[10]);
  Z[1][5]:=1/(thc[1]*thc[5])*(2*Y[5]*Y[1]*X[1]*X[5]+2*Y[8]*Y[7]*X[7]*X[8]-2*Y[3]*Y[13]*X[13]*X[3]-2*Y[9]*Y[14]*X[14]*X[9]+2*Y[6]*Y[4]*X[4]*X[6]+2*Y[16]*Y[11]*X[11]*X[16]-2*Y[15]*Y[2]*X[2]*X[15]-2*Y[10]*Y[12]*X[12]*X[10]);
  Z[1][6]:=1/(thc[1]*thc[6])*(2*Y[6]*Y[1]*X[1]*X[6]+2*Y[5]*Y[4]*X[4]*X[5]+2*Y[16]*Y[7]*X[7]*X[16]+2*Y[11]*Y[8]*X[8]*X[11]-2*Y[15]*Y[3]*X[3]*X[15]-2*Y[13]*Y[2]*X[2]*X[13]-2*Y[10]*Y[9]*X[9]*X[10]-2*Y[14]*Y[12]*X[12]*X[14]);
  Z[1][7]:=1/(thc[1]*thc[7])*(2*Y[1]*Y[7]*X[7]*X[1]+2*Y[5]*Y[8]*X[8]*X[5]+2*Y[9]*Y[3]*X[3]*X[9]+2*Y[14]*Y[13]*X[13]*X[14]-2*Y[11]*Y[4]*X[4]*X[11]-2*Y[16]*Y[6]*X[6]*X[16]-2*Y[12]*Y[2]*X[2]*X[12]-2*Y[15]*Y[10]*X[10]*X[15]);
  Z[1][8]:=1/(thc[1]*thc[8])*(2*Y[8]*Y[1]*X[1]*X[8]+2*Y[7]*Y[5]*X[5]*X[7]-2*Y[14]*Y[3]*X[3]*X[14]-2*Y[9]*Y[13]*X[13]*X[9]-2*Y[16]*Y[4]*X[4]*X[16]-2*Y[11]*Y[6]*X[6]*X[11]+2*Y[10]*Y[2]*X[2]*X[10]+2*Y[12]*Y[15]*X[15]*X[12]);
  Z[1][9]:=1/(thc[1]*thc[9])*(2*Y[9]*Y[1]*X[1]*X[9]+2*Y[14]*Y[5]*X[5]*X[14]+2*Y[3]*Y[7]*X[7]*X[3]+2*Y[13]*Y[8]*X[8]*X[13]-2*Y[12]*Y[4]*X[4]*X[12]-2*Y[10]*Y[6]*X[6]*X[10]-2*Y[2]*Y[11]*X[11]*X[2]-2*Y[15]*Y[16]*X[16]*X[15]);
  Z[1][10]:=1/(thc[1]*thc[10])*(2*Y[1]*Y[10]*X[10]*X[1]+2*Y[5]*Y[12]*X[12]*X[5]+2*Y[15]*Y[7]*X[7]*X[15]+2*Y[8]*Y[2]*X[2]*X[8]-2*Y[16]*Y[3]*X[3]*X[16]-2*Y[11]*Y[13]*X[13]*X[11]-2*Y[6]*Y[9]*X[9]*X[6]-2*Y[14]*Y[4]*X[4]*X[14]);
  Z[1][11]:=1/(thc[8]*thc[6])*(2*Y[11]*Y[1]*X[8]*X[6]+2*Y[16]*Y[5]*X[7]*X[4]+2*Y[4]*Y[7]*X[5]*X[16]+2*Y[6]*Y[8]*X[1]*X[11]+2*Y[12]*Y[3]*X[14]*X[15]+2*Y[10]*Y[13]*X[9]*X[2]+2*Y[2]*Y[9]*X[13]*X[10]+2*Y[15]*Y[14]*X[3]*X[12]);
  Z[1][12]:=1/(thc[7]*thc[2])*(2*Y[12]*Y[1]*X[7]*X[2]+2*Y[10]*Y[5]*X[8]*X[15]+2*Y[2]*Y[7]*X[1]*X[12]+2*Y[15]*Y[8]*X[5]*X[10]+2*Y[11]*Y[3]*X[9]*X[4]+2*Y[16]*Y[13]*X[14]*X[6]+2*Y[4]*Y[9]*X[3]*X[11]+2*Y[6]*Y[14]*X[13]*X[16]);
  Z[1][13]:=1/(thc[6]*thc[2])*(2*Y[13]*Y[1]*X[6]*X[2]+2*Y[3]*Y[5]*X[4]*X[15]- 2*Y[14]*Y[7]*X[16]*X[12]-2*Y[9]*Y[8]*X[11]*X[10]+2*Y[15]*Y[4]*X[5]*X[3]+2*Y[2]*Y[6]*X[1]*X[13]-2*Y[10]*Y[11]*X[8]*X[9]-2*Y[12]*Y[16]*X[7]*X[14]);
  Z[1][14]:=1/(thc[5]*thc[9])*(2*Y[14]*Y[1]*X[5]*X[9]+2*Y[9]*Y[5]*X[1]*X[14]+2*Y[13]*Y[7]*X[8]*X[3]+2*Y[3]*Y[8]*X[7]*X[13]-2*Y[10]*Y[4]*X[6]*X[12]-2*Y[12]*Y[6]*X[4]*X[10]-2*Y[15]*Y[11]*X[16]*X[2]-2*Y[2]*Y[16]*X[11]*X[15]);
  Z[1][15]:=1/(thc[5]*thc[2])*(2*Y[15]*Y[1]*X[5]*X[2]+2*Y[2]*Y[5]*X[1]*X[15]+2*Y[10]*Y[7]*X[8]*X[12]+2*Y[12]*Y[8]*X[7]*X[10]+2*Y[6]*Y[3]*X[13]*X[4]+2*Y[4]*Y[13]*X[3]*X[6]+2*Y[16]*Y[9]*X[14]*X[11]+2*Y[11]*Y[14]*X[9]*X[16]);
  Z[1][16]:=1/(thc[3]*thc[10])*(2*Y[16]*Y[1]*X[3]*X[10]-2*Y[11]*Y[5]*X[13]*X[12]+2*Y[6]*Y[7]*X[9]*X[15]-2*Y[4]*Y[8]*X[14]*X[2]+2*Y[10]*Y[3]*X[1]*X[16]-2*Y[12]*Y[13]*X[5]*X[11]+2*Y[15]*Y[9]*X[7]*X[6]-2*Y[2]*Y[14]*X[8]*X[4]);

  Z[2][1]:=1/(thc[2]*thc[1])*(2*Y[1]*Y[2]*X[2]*X[1]-2*Y[5]*Y[15]*X[15]*X[5]-2*Y[7]*Y[12]*X[12]*X[7]+2*Y[8]*Y[10]*X[10]*X[8]+2*Y[3]*Y[4]*X[4]*X[3]-2*Y[13]*Y[6]*X[6]*X[13]-2*Y[9]*Y[11]*X[11]*X[9]+2*Y[14]*Y[16]*X[16]*X[14]);
  Z[2][2]:=1/(thc[2]^2)*(+Y[2]^2*X[2]^2-Y[15]^2*X[15]^2-Y[12]^2*X[12]^2+Y[10]^2*X[10]^2+Y[4]^2*X[4]^2-Y[6]^2*X[6]^2-Y[11]^2*X[11]^2+Y[16]^2*X[16]^2+Y[3]^2*X[3]^2-Y[13]^2*X[13]^2-Y[9]^2*X[9]^2+Y[14]^2*X[14]^2+Y[1]^2*X[1]^2-Y[5]^2*X[5]^2-Y[7]^2*X[7]^2+Y[8]^2*X[8]^2);
  Z[2][3]:=1/(thc[2]*thc[3])*(2*Y[3]*Y[2]*X[2]*X[3]-2*Y[13]*Y[15]*X[15]*X[13]-2*Y[9]*Y[12]*X[12]*X[9]+2*Y[14]*Y[10]*X[10]*X[14]+2*Y[1]*Y[4]*X[4]*X[1]-2*Y[5]*Y[6]*X[6]*X[5]-2*Y[7]*Y[11]*X[11]*X[7]+2*Y[8]*Y[16]*X[16]*X[8]);
  Z[2][4]:=1/(thc[2]*thc[4])*(2*Y[4]*Y[2]*X[2]*X[4]-2*Y[6]*Y[15]*X[15]*X[6]-2*Y[11]*Y[12]*X[12]*X[11]+2*Y[16]*Y[10]*X[10]*X[16]+2*Y[1]*Y[3]*X[3]*X[1]-2*Y[5]*Y[13]*X[13]*X[5]-2*Y[7]*Y[9]*X[9]*X[7]+2*Y[8]*Y[14]*X[14]*X[8]);
  Z[2][5]:=1/(thc[2]*thc[5])*(2*Y[5]*Y[2]*X[2]*X[5]-2*Y[1]*Y[15]*X[15]*X[1]-2*Y[8]*Y[12]*X[12]*X[8]+2*Y[7]*Y[10]*X[10]*X[7]-2*Y[13]*Y[4]*X[4]*X[13]+2*Y[3]*Y[6]*X[6]*X[3]+2*Y[14]*Y[11]*X[11]*X[14]-2*Y[9]*Y[16]*X[16]*X[9]);
  Z[2][6]:=1/(thc[2]*thc[6])*(2*Y[6]*Y[2]*X[2]*X[6]-2*Y[4]*Y[15]*X[15]*X[4]-2*Y[16]*Y[12]*X[12]*X[16]+2*Y[11]*Y[10]*X[10]*X[11]+2*Y[5]*Y[3]*X[3]*X[5]-2*Y[1]*Y[13]*X[13]*X[1]-2*Y[8]*Y[9]*X[9]*X[8]+2*Y[7]*Y[14]*X[14]*X[7]);
  Z[2][7]:=1/(thc[2]*thc[7])*(2*Y[7]*Y[2]*X[2]*X[7]-2*Y[8]*Y[15]*X[15]*X[8]-2*Y[1]*Y[12]*X[12]*X[1]+2*Y[5]*Y[10]*X[10]*X[5]+2*Y[9]*Y[4]*X[4]*X[9]-2*Y[14]*Y[6]*X[6]*X[14]-2*Y[3]*Y[11]*X[11]*X[3]+2*Y[13]*Y[16]*X[16]*X[13]);
  Z[2][8]:=1/(thc[2]*thc[8])*(2*Y[8]*Y[2]*X[2]*X[8]-2*Y[7]*Y[15]*X[15]*X[7]-2*Y[5]*Y[12]*X[12]*X[5]+2*Y[1]*Y[10]*X[10]*X[1]-2*Y[14]*Y[4]*X[4]*X[14]+2*Y[9]*Y[6]*X[6]*X[9]+2*Y[13]*Y[11]*X[11]*X[13]-2*Y[3]*Y[16]*X[16]*X[3]);
  Z[2][9]:=1/(thc[2]*thc[9])*(2*Y[9]*Y[2]*X[2]*X[9]-2*Y[14]*Y[15]*X[15]*X[14]-2*Y[3]*Y[12]*X[12]*X[3]+2*Y[13]*Y[10]*X[10]*X[13]+2*Y[7]*Y[4]*X[4]*X[7]-2*Y[8]*Y[6]*X[6]*X[8]-2*Y[1]*Y[11]*X[11]*X[1]+2*Y[5]*Y[16]*X[16]*X[5]);
  Z[2][10]:=1/(thc[2]*thc[10])*(2*Y[10]*Y[2]*X[2]*X[10]-2*Y[12]*Y[15]*X[15]*X[12]-2*Y[16]*Y[4]*X[4]*X[16]+2*Y[11]*Y[6]*X[6]*X[11]-2*Y[14]*Y[3]*X[3]*X[14]+2*Y[9]*Y[13]*X[13]*X[9]+2*Y[8]*Y[1]*X[1]*X[8]-2*Y[7]*Y[5]*X[5]*X[7]);
  Z[2][11]:=1/(thc[10]*thc[6])*(2*Y[11]*Y[2]*X[10]*X[6]-2*Y[16]*Y[15]*X[12]*X[4]-2*Y[4]*Y[12]*X[15]*X[16]+2*Y[6]*Y[10]*X[2]*X[11]-2*Y[7]*Y[3]*X[14]*X[5]+2*Y[8]*Y[13]*X[9]*X[1]+2*Y[1]*Y[9]*X[13]*X[8]-2*Y[5]*Y[14]*X[3]*X[7]);
  Z[2][12]:=1/(thc[3]*thc[9])*(2*Y[12]*Y[2]*X[3]*X[9]-2*Y[10]*Y[15]*X[13]*X[14]+2*Y[11]*Y[4]*X[1]*X[7]-2*Y[16]*Y[6]*X[5]*X[8]+2*Y[9]*Y[3]*X[2]*X[12]-2*Y[14]*Y[13]*X[15]*X[10]+2*Y[7]*Y[1]*X[4]*X[11]-2*Y[8]*Y[5]*X[6]*X[16]);
  Z[2][13]:=1/(thc[10]*thc[9])*(2*Y[13]*Y[2]*X[10]*X[9]-2*Y[3]*Y[15]*X[12]*X[14]-2*Y[14]*Y[12]*X[15]*X[3]+2*Y[9]*Y[10]*X[2]*X[13]-2*Y[5]*Y[4]*X[16]*X[7]+2*Y[1]*Y[6]*X[11]*X[8]+2*Y[8]*Y[11]*X[6]*X[1]-2*Y[7]*Y[16]*X[4]*X[5]);
  Z[2][14]:=1/(thc[10]*thc[3])*(2*Y[14]*Y[2]*X[10]*X[3]-2*Y[9]*Y[15]*X[12]*X[13]-2*Y[13]*Y[12]*X[15]*X[9]+2*Y[3]*Y[10]*X[2]*X[14]-2*Y[8]*Y[4]*X[16]*X[1]+2*Y[7]*Y[6]*X[11]*X[5]+2*Y[5]*Y[11]*X[6]*X[7]-2*Y[1]*Y[16]*X[4]*X[8]);
  Z[2][15]:=1/(thc[4]*thc[6])*(2*Y[15]*Y[2]*X[4]*X[6]-2*Y[10]*Y[12]*X[11]*X[16]+2*Y[6]*Y[4]*X[2]*X[15]-2*Y[16]*Y[11]*X[12]*X[10]+2*Y[13]*Y[3]*X[1]*X[5]-2*Y[14]*Y[9]*X[7]*X[8]+2*Y[5]*Y[1]*X[3]*X[13]-2*Y[8]*Y[7]*X[9]*X[14]);
  Z[2][16]:=1/(thc[9]*thc[5])*(2*Y[16]*Y[2]*X[9]*X[5]-2*Y[11]*Y[15]*X[14]*X[1]+2*Y[6]*Y[12]*X[3]*X[8]-2*Y[4]*Y[10]*X[13]*X[7]+2*Y[8]*Y[3]*X[12]*X[6]-2*Y[7]*Y[13]*X[10]*X[4]+2*Y[5]*Y[9]*X[2]*X[16]-2*Y[1]*Y[14]*X[15]*X[11]);

  Z[3][1]:=1/(thc[3]*thc[1])*(2*Y[1]*Y[3]*X[3]*X[1]-2*Y[5]*Y[13]*X[13]*X[5]+2*Y[7]*Y[9]*X[9]*X[7]-2*Y[8]*Y[14]*X[14]*X[8]+2*Y[4]*Y[2]*X[2]*X[4]-2*Y[6]*Y[15]*X[15]*X[6]+2*Y[11]*Y[12]*X[12]*X[11]-2*Y[16]*Y[10]*X[10]*X[16]);
  Z[3][2]:=1/(thc[3]*thc[2])*(2*Y[2]*Y[3]*X[3]*X[2]-2*Y[15]*Y[13]*X[13]*X[15]+2*Y[12]*Y[9]*X[9]*X[12]-2*Y[10]*Y[14]*X[14]*X[10]+2*Y[4]*Y[1]*X[1]*X[4]-2*Y[6]*Y[5]*X[5]*X[6]+2*Y[11]*Y[7]*X[7]*X[11]-2*Y[16]*Y[8]*X[8]*X[16]);
  Z[3][3]:=1/(thc[3]^2)*(+Y[3]^2*X[3]^2-Y[13]^2*X[13]^2+Y[9]^2*X[9]^2-Y[14]^2*X[14]^2+Y[1]^2*X[1]^2-Y[5]^2*X[5]^2+Y[7]^2*X[7]^2-Y[8]^2*X[8]^2+Y[2]^2*X[2]^2-Y[15]^2*X[15]^2+Y[12]^2*X[12]^2-Y[10]^2*X[10]^2+Y[4]^2*X[4]^2-Y[6]^2*X[6]^2+Y[11]^2*X[11]^2-Y[16]^2*X[16]^2);
  Z[3][4]:=1/(thc[3]*thc[4])*(2*Y[4]*Y[3]*X[3]*X[4]-2*Y[6]*Y[13]*X[13]*X[6]+2*Y[11]*Y[9]*X[9]*X[11]-2*Y[16]*Y[14]*X[14]*X[16]+2*Y[2]*Y[1]*X[1]*X[2]-2*Y[15]*Y[5]*X[5]*X[15]+2*Y[12]*Y[7]*X[7]*X[12]-2*Y[10]*Y[8]*X[8]*X[10]);
  Z[3][5]:=1/(thc[3]*thc[5])*(2*Y[5]*Y[3]*X[3]*X[5]-2*Y[1]*Y[13]*X[13]*X[1]+2*Y[8]*Y[9]*X[9]*X[8]-2*Y[7]*Y[14]*X[14]*X[7]+2*Y[6]*Y[2]*X[2]*X[6]-2*Y[4]*Y[15]*X[15]*X[4]+2*Y[16]*Y[12]*X[12]*X[16]-2*Y[11]*Y[10]*X[10]*X[11]);
  Z[3][6]:=1/(thc[3]*thc[6])*(2*Y[6]*Y[3]*X[3]*X[6]-2*Y[4]*Y[13]*X[13]*X[4]+2*Y[16]*Y[9]*X[9]*X[16]-2*Y[11]*Y[14]*X[14]*X[11]-2*Y[15]*Y[1]*X[1]*X[15]+2*Y[2]*Y[5]*X[5]*X[2]-2*Y[10]*Y[7]*X[7]*X[10]+2*Y[12]*Y[8]*X[8]*X[12]);
  Z[3][7]:=1/(thc[3]*thc[7])*(2*Y[7]*Y[3]*X[3]*X[7]-2*Y[8]*Y[13]*X[13]*X[8]+2*Y[1]*Y[9]*X[9]*X[1]-2*Y[5]*Y[14]*X[14]*X[5]-2*Y[11]*Y[2]*X[2]*X[11]+2*Y[16]*Y[15]*X[15]*X[16]-2*Y[4]*Y[12]*X[12]*X[4]+2*Y[6]*Y[10]*X[10]*X[6]);
  Z[3][8]:=1/(thc[3]*thc[8])*(2*Y[8]*Y[3]*X[3]*X[8]-2*Y[7]*Y[13]*X[13]*X[7]+2*Y[5]*Y[9]*X[9]*X[5]-2*Y[1]*Y[14]*X[14]*X[1]-2*Y[16]*Y[2]*X[2]*X[16]+2*Y[11]*Y[15]*X[15]*X[11]-2*Y[6]*Y[12]*X[12]*X[6]+2*Y[4]*Y[10]*X[10]*X[4]);
  Z[3][9]:=1/(thc[3]*thc[9])*(2*Y[9]*Y[3]*X[3]*X[9]-2*Y[14]*Y[13]*X[13]*X[14]+2*Y[7]*Y[1]*X[1]*X[7]-2*Y[8]*Y[5]*X[5]*X[8]-2*Y[12]*Y[2]*X[2]*X[12]+2*Y[10]*Y[15]*X[15]*X[10]-2*Y[11]*Y[4]*X[4]*X[11]+2*Y[16]*Y[6]*X[6]*X[16]);
  Z[3][10]:=1/(thc[3]*thc[10])*(2*Y[10]*Y[3]*X[3]*X[10]-2*Y[12]*Y[13]*X[13]*X[12]+2*Y[15]*Y[9]*X[9]*X[15]-2*Y[2]*Y[14]*X[14]*X[2]-2*Y[16]*Y[1]*X[1]*X[16]+2*Y[11]*Y[5]*X[5]*X[11]-2*Y[6]*Y[7]*X[7]*X[6]+2*Y[4]*Y[8]*X[8]*X[4]);
  Z[3][11]:=1/(thc[5]*thc[10])*(2*Y[11]*Y[3]*X[5]*X[10]-2*Y[16]*Y[13]*X[1]*X[12]+2*Y[4]*Y[9]*X[8]*X[15]-2*Y[6]*Y[14]*X[7]*X[2]-2*Y[12]*Y[1]*X[13]*X[16]+2*Y[10]*Y[5]*X[3]*X[11]-2*Y[2]*Y[7]*X[14]*X[6]+2*Y[15]*Y[8]*X[9]*X[4]);
  Z[3][12]:=1/(thc[7]*thc[4])*(2*Y[12]*Y[3]*X[7]*X[4]+2*Y[10]*Y[13]*X[8]*X[6]+2*Y[2]*Y[9]*X[1]*X[11]+2*Y[15]*Y[14]*X[5]*X[16]+2*Y[11]*Y[1]*X[9]*X[2]+2*Y[16]*Y[5]*X[14]*X[15]+2*Y[4]*Y[7]*X[3]*X[12]+2*Y[6]*Y[8]*X[13]*X[10]);
  Z[3][13]:=1/(thc[1]*thc[5])*(2*Y[13]*Y[3]*X[1]*X[5]+2*Y[14]*Y[9]*X[7]*X[8]+2*Y[5]*Y[1]*X[3]*X[13]+2*Y[8]*Y[7]*X[9]*X[14]+2*Y[15]*Y[2]*X[4]*X[6]+2*Y[10]*Y[12]*X[11]*X[16]+2*Y[6]*Y[4]*X[2]*X[15]+2*Y[16]*Y[11]*X[12]*X[10]);
  Z[3][14]:=1/(thc[1]*thc[8])*(2*Y[14]*Y[3]*X[1]*X[8]+2*Y[9]*Y[13]*X[5]*X[7]+2*Y[8]*Y[1]*X[3]*X[14]+2*Y[7]*Y[5]*X[13]*X[9]-2*Y[10]*Y[2]*X[4]*X[16]-2*Y[12]*Y[15]*X[6]*X[11]-2*Y[16]*Y[4]*X[2]*X[10]-2*Y[11]*Y[6]*X[15]*X[12]);
  Z[3][15]:=1/(thc[1]*thc[6])*(2*Y[15]*Y[3]*X[1]*X[6]+2*Y[2]*Y[13]*X[5]*X[4]+2*Y[10]*Y[9]*X[7]*X[16]+2*Y[12]*Y[14]*X[8]*X[11]+2*Y[6]*Y[1]*X[3]*X[15]+2*Y[4]*Y[5]*X[13]*X[2]+2*Y[16]*Y[7]*X[9]*X[10]+2*Y[11]*Y[8]*X[14]*X[12]);
  Z[3][16]:=1/(thc[8]*thc[2])*(2*Y[16]*Y[3]*X[8]*X[2]-2*Y[11]*Y[13]*X[7]*X[15]+2*Y[6]*Y[9]*X[5]*X[12]-2*Y[4]*Y[14]*X[1]*X[10]-2*Y[10]*Y[1]*X[14]*X[4]+2*Y[12]*Y[5]*X[9]*X[6]-2*Y[15]*Y[7]*X[13]*X[11]+2*Y[2]*Y[8]*X[3]*X[16]);

  Z[4][1]:=1/(thc[4]*thc[1])*(2*Y[1]*Y[4]*X[4]*X[1]+2*Y[5]*Y[6]*X[6]*X[5]-2*Y[7]*Y[11]*X[11]*X[7]-2*Y[8]*Y[16]*X[16]*X[8]+2*Y[3]*Y[2]*X[2]*X[3]+2*Y[13]*Y[15]*X[15]*X[13]-2*Y[9]*Y[12]*X[12]*X[9]-2*Y[14]*Y[10]*X[10]*X[14]);
  Z[4][2]:=1/(thc[4]*thc[2])*(2*Y[2]*Y[4]*X[4]*X[2]+2*Y[15]*Y[6]*X[6]*X[15]-2*Y[12]*Y[11]*X[11]*X[12]-2*Y[10]*Y[16]*X[16]*X[10]+2*Y[3]*Y[1]*X[1]*X[3]+2*Y[13]*Y[5]*X[5]*X[13]-2*Y[9]*Y[7]*X[7]*X[9]-2*Y[14]*Y[8]*X[8]*X[14]);
  Z[4][3]:=1/(thc[4]*thc[3])*(2*Y[3]*Y[4]*X[4]*X[3]+2*Y[13]*Y[6]*X[6]*X[13]-2*Y[9]*Y[11]*X[11]*X[9]-2*Y[14]*Y[16]*X[16]*X[14]+2*Y[1]*Y[2]*X[2]*X[1]+2*Y[5]*Y[15]*X[15]*X[5]-2*Y[7]*Y[12]*X[12]*X[7]-2*Y[8]*Y[10]*X[10]*X[8]);
  Z[4][4]:=1/(thc[4]^2)*(+Y[4]^2*X[4]^2+Y[6]^2*X[6]^2-Y[11]^2*X[11]^2-Y[16]^2*X[16]^2+Y[2]^2*X[2]^2+Y[15]^2*X[15]^2-Y[12]^2*X[12]^2-Y[10]^2*X[10]^2+Y[1]^2*X[1]^2+Y[5]^2*X[5]^2-Y[7]^2*X[7]^2-Y[8]^2*X[8]^2+Y[3]^2*X[3]^2+Y[13]^2*X[13]^2-Y[9]^2*X[9]^2-Y[14]^2*X[14]^2);
  Z[4][5]:=1/(thc[4]*thc[5])*(2*Y[5]*Y[4]*X[4]*X[5]+2*Y[1]*Y[6]*X[6]*X[1]-2*Y[8]*Y[11]*X[11]*X[8]-2*Y[7]*Y[16]*X[16]*X[7]-2*Y[13]*Y[2]*X[2]*X[13]-2*Y[3]*Y[15]*X[15]*X[3]+2*Y[14]*Y[12]*X[12]*X[14]+2*Y[9]*Y[10]*X[10]*X[9]);
  Z[4][6]:=1/(thc[4]*thc[6])*(2*Y[6]*Y[4]*X[4]*X[6]-2*Y[16]*Y[11]*X[11]*X[16]-2*Y[15]*Y[2]*X[2]*X[15]+2*Y[10]*Y[12]*X[12]*X[10]+2*Y[5]*Y[1]*X[1]*X[5]-2*Y[8]*Y[7]*X[7]*X[8]-2*Y[13]*Y[3]*X[3]*X[13]+2*Y[14]*Y[9]*X[9]*X[14]);
  Z[4][7]:=1/(thc[4]*thc[7])*(2*Y[7]*Y[4]*X[4]*X[7]+2*Y[8]*Y[6]*X[6]*X[8]-2*Y[1]*Y[11]*X[11]*X[1]-2*Y[5]*Y[16]*X[16]*X[5]+2*Y[9]*Y[2]*X[2]*X[9]+2*Y[14]*Y[15]*X[15]*X[14]-2*Y[3]*Y[12]*X[12]*X[3]-2*Y[13]*Y[10]*X[10]*X[13]);
  Z[4][8]:=1/(thc[4]*thc[8])*(2*Y[8]*Y[4]*X[4]*X[8]+2*Y[7]*Y[6]*X[6]*X[7]-2*Y[5]*Y[11]*X[11]*X[5]-2*Y[1]*Y[16]*X[16]*X[1]-2*Y[14]*Y[2]*X[2]*X[14]-2*Y[9]*Y[15]*X[15]*X[9]+2*Y[13]*Y[12]*X[12]*X[13]+2*Y[3]*Y[10]*X[10]*X[3]);
  Z[4][9]:=1/(thc[4]*thc[9])*(2*Y[9]*Y[4]*X[4]*X[9]+2*Y[14]*Y[6]*X[6]*X[14]-2*Y[3]*Y[11]*X[11]*X[3]-2*Y[13]*Y[16]*X[16]*X[13]+2*Y[7]*Y[2]*X[2]*X[7]+2*Y[8]*Y[15]*X[15]*X[8]-2*Y[1]*Y[12]*X[12]*X[1]-2*Y[5]*Y[10]*X[10]*X[5]);
  Z[4][10]:=1/(thc[4]*thc[10])*(2*Y[10]*Y[4]*X[4]*X[10]+2*Y[12]*Y[6]*X[6]*X[12]-2*Y[15]*Y[11]*X[11]*X[15]-2*Y[2]*Y[16]*X[16]*X[2]-2*Y[14]*Y[1]*X[1]*X[14]-2*Y[9]*Y[5]*X[5]*X[9]+2*Y[13]*Y[7]*X[7]*X[13]+2*Y[3]*Y[8]*X[8]*X[3]);
  Z[4][11]:=1/(thc[1]*thc[7])*(2*Y[11]*Y[4]*X[1]*X[7]+2*Y[16]*Y[6]*X[5]*X[8]+2*Y[12]*Y[2]*X[3]*X[9]+2*Y[10]*Y[15]*X[13]*X[14]+2*Y[7]*Y[1]*X[4]*X[11]+2*Y[8]*Y[5]*X[6]*X[16]+2*Y[9]*Y[3]*X[2]*X[12]+2*Y[14]*Y[13]*X[15]*X[10]);
  Z[4][12]:=1/(thc[1]*thc[9])*(2*Y[12]*Y[4]*X[1]*X[9]+2*Y[10]*Y[6]*X[5]*X[14]+2*Y[2]*Y[11]*X[7]*X[3]+2*Y[15]*Y[16]*X[8]*X[13]+2*Y[9]*Y[1]*X[4]*X[12]+2*Y[14]*Y[5]*X[6]*X[10]+2*Y[3]*Y[7]*X[11]*X[2]+2*Y[13]*Y[8]*X[16]*X[15]);
  Z[4][13]:=1/(thc[6]*thc[3])*(2*Y[13]*Y[4]*X[6]*X[3]+2*Y[3]*Y[6]*X[4]*X[13]-2*Y[14]*Y[11]*X[16]*X[9]-2*Y[9]*Y[16]*X[11]*X[14]+2*Y[5]*Y[2]*X[15]*X[1]+2*Y[1]*Y[15]*X[2]*X[5]-2*Y[8]*Y[12]*X[10]*X[7]-2*Y[7]*Y[10]*X[12]*X[8]);
  Z[4][14]:=1/(thc[10]*thc[1])*(2*Y[14]*Y[4]*X[10]*X[1]+2*Y[9]*Y[6]*X[12]*X[5]-2*Y[13]*Y[11]*X[15]*X[7]-2*Y[3]*Y[16]*X[2]*X[8]-2*Y[8]*Y[2]*X[16]*X[3]-2*Y[7]*Y[15]*X[11]*X[13]+2*Y[5]*Y[12]*X[6]*X[9]+2*Y[1]*Y[10]*X[4]*X[14]);
  Z[4][15]:=1/(thc[6]*thc[2])*(2*Y[15]*Y[4]*X[6]*X[2]+2*Y[2]*Y[6]*X[4]*X[15]-2*Y[10]*Y[11]*X[16]*X[12]-2*Y[12]*Y[16]*X[11]*X[10]+2*Y[13]*Y[1]*X[5]*X[3]+2*Y[3]*Y[5]*X[1]*X[13]-2*Y[14]*Y[7]*X[8]*X[9]-2*Y[9]*Y[8]*X[7]*X[14]);
  Z[4][16]:=1/(thc[5]*thc[7])*(2*Y[16]*Y[4]*X[5]*X[7]+2*Y[11]*Y[6]*X[1]*X[8]-2*Y[10]*Y[2]*X[13]*X[9]-2*Y[12]*Y[15]*X[3]*X[14]+2*Y[8]*Y[1]*X[6]*X[11]+2*Y[7]*Y[5]*X[4]*X[16]-2*Y[14]*Y[3]*X[15]*X[12]-2*Y[9]*Y[13]*X[2]*X[10]);

  Z[5][1]:=1/(thc[5]*thc[1])*(2*Y[1]*Y[5]*X[5]*X[1]+2*Y[7]*Y[8]*X[8]*X[7]+2*Y[3]*Y[13]*X[13]*X[3]+2*Y[9]*Y[14]*X[14]*X[9]+2*Y[4]*Y[6]*X[6]*X[4]+2*Y[11]*Y[16]*X[16]*X[11]+2*Y[2]*Y[15]*X[15]*X[2]+2*Y[12]*Y[10]*X[10]*X[12]);
  Z[5][2]:=1/(thc[5]*thc[2])*(2*Y[2]*Y[5]*X[5]*X[2]+2*Y[15]*Y[1]*X[1]*X[15]+2*Y[12]*Y[8]*X[8]*X[12]+2*Y[10]*Y[7]*X[7]*X[10]+2*Y[4]*Y[13]*X[13]*X[4]+2*Y[6]*Y[3]*X[3]*X[6]+2*Y[11]*Y[14]*X[14]*X[11]+2*Y[16]*Y[9]*X[9]*X[16]);
  Z[5][3]:=1/(thc[5]*thc[3])*(2*Y[3]*Y[5]*X[5]*X[3]+2*Y[13]*Y[1]*X[1]*X[13]+2*Y[9]*Y[8]*X[8]*X[9]+2*Y[14]*Y[7]*X[7]*X[14]+2*Y[2]*Y[6]*X[6]*X[2]+2*Y[15]*Y[4]*X[4]*X[15]+2*Y[12]*Y[16]*X[16]*X[12]+2*Y[10]*Y[11]*X[11]*X[10]);
  Z[5][4]:=1/(thc[5]*thc[4])*(2*Y[4]*Y[5]*X[5]*X[4]+2*Y[6]*Y[1]*X[1]*X[6]+2*Y[11]*Y[8]*X[8]*X[11]+2*Y[16]*Y[7]*X[7]*X[16]+2*Y[2]*Y[13]*X[13]*X[2]+2*Y[15]*Y[3]*X[3]*X[15]+2*Y[12]*Y[14]*X[14]*X[12]+2*Y[10]*Y[9]*X[9]*X[10]);
  Z[5][5]:=1/(thc[5]^2)*(+Y[5]^2*X[5]^2+Y[1]^2*X[1]^2+Y[8]^2*X[8]^2+Y[7]^2*X[7]^2-Y[13]^2*X[13]^2-Y[3]^2*X[3]^2-Y[14]^2*X[14]^2-Y[9]^2*X[9]^2+Y[6]^2*X[6]^2+Y[4]^2*X[4]^2+Y[16]^2*X[16]^2+Y[11]^2*X[11]^2-Y[15]^2*X[15]^2-Y[2]^2*X[2]^2-Y[10]^2*X[10]^2-Y[12]^2*X[12]^2);
  Z[5][6]:=1/(thc[5]*thc[6])*(2*Y[6]*Y[5]*X[5]*X[6]+2*Y[4]*Y[1]*X[1]*X[4]+2*Y[16]*Y[8]*X[8]*X[16]+2*Y[11]*Y[7]*X[7]*X[11]-2*Y[15]*Y[13]*X[13]*X[15]-2*Y[2]*Y[3]*X[3]*X[2]-2*Y[10]*Y[14]*X[14]*X[10]-2*Y[12]*Y[9]*X[9]*X[12]);
  Z[5][7]:=1/(thc[5]*thc[7])*(2*Y[7]*Y[5]*X[5]*X[7]+2*Y[8]*Y[1]*X[1]*X[8]+2*Y[9]*Y[13]*X[13]*X[9]+2*Y[14]*Y[3]*X[3]*X[14]-2*Y[11]*Y[6]*X[6]*X[11]-2*Y[16]*Y[4]*X[4]*X[16]-2*Y[12]*Y[15]*X[15]*X[12]-2*Y[10]*Y[2]*X[2]*X[10]);
  Z[5][8]:=1/(thc[5]*thc[8])*(2*Y[8]*Y[5]*X[5]*X[8]+2*Y[7]*Y[1]*X[1]*X[7]-2*Y[14]*Y[13]*X[13]*X[14]-2*Y[9]*Y[3]*X[3]*X[9]-2*Y[16]*Y[6]*X[6]*X[16]-2*Y[11]*Y[4]*X[4]*X[11]+2*Y[10]*Y[15]*X[15]*X[10]+2*Y[12]*Y[2]*X[2]*X[12]);
  Z[5][9]:=1/(thc[5]*thc[9])*(2*Y[9]*Y[5]*X[5]*X[9]+2*Y[14]*Y[1]*X[1]*X[14]+2*Y[3]*Y[8]*X[8]*X[3]+2*Y[13]*Y[7]*X[7]*X[13]-2*Y[12]*Y[6]*X[6]*X[12]-2*Y[10]*Y[4]*X[4]*X[10]-2*Y[2]*Y[16]*X[16]*X[2]-2*Y[15]*Y[11]*X[11]*X[15]);
  Z[5][10]:=1/(thc[5]*thc[10])*(2*Y[10]*Y[5]*X[5]*X[10]+2*Y[12]*Y[1]*X[1]*X[12]+2*Y[15]*Y[8]*X[8]*X[15]+2*Y[2]*Y[7]*X[7]*X[2]-2*Y[16]*Y[13]*X[13]*X[16]-2*Y[11]*Y[3]*X[3]*X[11]-2*Y[6]*Y[14]*X[14]*X[6]-2*Y[4]*Y[9]*X[9]*X[4]);
  Z[5][11]:=1/(thc[8]*thc[4])*(2*Y[11]*Y[5]*X[8]*X[4]+2*Y[16]*Y[1]*X[7]*X[6]+2*Y[4]*Y[8]*X[5]*X[11]+2*Y[6]*Y[7]*X[1]*X[16]+2*Y[12]*Y[13]*X[14]*X[2]+2*Y[10]*Y[3]*X[9]*X[15]+2*Y[2]*Y[14]*X[13]*X[12]+2*Y[15]*Y[9]*X[3]*X[10]);
  Z[5][12]:=1/(thc[1]*thc[10])*(2*Y[12]*Y[5]*X[1]*X[10]+2*Y[10]*Y[1]*X[5]*X[12]+2*Y[2]*Y[8]*X[7]*X[15]+2*Y[15]*Y[7]*X[8]*X[2]-2*Y[11]*Y[13]*X[3]*X[16]-2*Y[16]*Y[3]*X[13]*X[11]-2*Y[4]*Y[14]*X[9]*X[6]-2*Y[6]*Y[9]*X[14]*X[4]);
  Z[5][13]:=1/(thc[4]*thc[2])*(2*Y[13]*Y[5]*X[4]*X[2]+2*Y[3]*Y[1]*X[6]*X[15]-2*Y[14]*Y[8]*X[11]*X[12]-2*Y[9]*Y[7]*X[16]*X[10]+2*Y[15]*Y[6]*X[1]*X[3]+2*Y[2]*Y[4]*X[5]*X[13]-2*Y[10]*Y[16]*X[7]*X[9]-2*Y[12]*Y[11]*X[8]*X[14]);
  Z[5][14]:=1/(thc[1]*thc[9])*(2*Y[14]*Y[5]*X[1]*X[9]+2*Y[9]*Y[1]*X[5]*X[14]+2*Y[13]*Y[8]*X[7]*X[3]+2*Y[3]*Y[7]*X[8]*X[13]-2*Y[10]*Y[6]*X[4]*X[12]-2*Y[12]*Y[4]*X[6]*X[10]-2*Y[15]*Y[16]*X[11]*X[2]-2*Y[2]*Y[11]*X[16]*X[15]);
  Z[5][15]:=1/(thc[8]*thc[10])*(2*Y[15]*Y[5]*X[8]*X[10]+2*Y[2]*Y[1]*X[7]*X[12]+2*Y[10]*Y[8]*X[5]*X[15]+2*Y[12]*Y[7]*X[1]*X[2]-2*Y[6]*Y[13]*X[14]*X[16]-2*Y[4]*Y[3]*X[9]*X[11]-2*Y[16]*Y[14]*X[13]*X[6]-2*Y[11]*Y[9]*X[3]*X[4]);
  Z[5][16]:=1/(thc[9]*thc[2])*(2*Y[16]*Y[5]*X[9]*X[2]-2*Y[11]*Y[1]*X[14]*X[15]+2*Y[6]*Y[8]*X[3]*X[12]-2*Y[4]*Y[7]*X[13]*X[10]-2*Y[10]*Y[13]*X[7]*X[4]+2*Y[12]*Y[3]*X[8]*X[6]-2*Y[15]*Y[14]*X[1]*X[11]+2*Y[2]*Y[9]*X[5]*X[16]);

  Z[6][1]:=1/(thc[6]*thc[1])*(2*Y[1]*Y[6]*X[6]*X[1]+2*Y[5]*Y[4]*X[4]*X[5]-2*Y[7]*Y[16]*X[16]*X[7]-2*Y[8]*Y[11]*X[11]*X[8]+2*Y[3]*Y[15]*X[15]*X[3]+2*Y[13]*Y[2]*X[2]*X[13]-2*Y[9]*Y[10]*X[10]*X[9]-2*Y[14]*Y[12]*X[12]*X[14]);
  Z[6][2]:=1/(thc[6]*thc[2])*(2*Y[2]*Y[6]*X[6]*X[2]+2*Y[15]*Y[4]*X[4]*X[15]-2*Y[12]*Y[16]*X[16]*X[12]-2*Y[10]*Y[11]*X[11]*X[10]+2*Y[3]*Y[5]*X[5]*X[3]+2*Y[13]*Y[1]*X[1]*X[13]-2*Y[9]*Y[8]*X[8]*X[9]-2*Y[14]*Y[7]*X[7]*X[14]);
  Z[6][3]:=1/(thc[6]*thc[3])*(2*Y[3]*Y[6]*X[6]*X[3]+2*Y[13]*Y[4]*X[4]*X[13]-2*Y[9]*Y[16]*X[16]*X[9]-2*Y[14]*Y[11]*X[11]*X[14]+2*Y[1]*Y[15]*X[15]*X[1]+2*Y[5]*Y[2]*X[2]*X[5]-2*Y[7]*Y[10]*X[10]*X[7]-2*Y[8]*Y[12]*X[12]*X[8]);
  Z[6][4]:=1/(thc[6]*thc[4])*(2*Y[4]*Y[6]*X[6]*X[4]-2*Y[11]*Y[16]*X[16]*X[11]+2*Y[2]*Y[15]*X[15]*X[2]-2*Y[12]*Y[10]*X[10]*X[12]+2*Y[1]*Y[5]*X[5]*X[1]-2*Y[7]*Y[8]*X[8]*X[7]+2*Y[3]*Y[13]*X[13]*X[3]-2*Y[9]*Y[14]*X[14]*X[9]);
  Z[6][5]:=1/(thc[6]*thc[5])*(2*Y[5]*Y[6]*X[6]*X[5]+2*Y[1]*Y[4]*X[4]*X[1]-2*Y[8]*Y[16]*X[16]*X[8]-2*Y[7]*Y[11]*X[11]*X[7]-2*Y[13]*Y[15]*X[15]*X[13]-2*Y[3]*Y[2]*X[2]*X[3]+2*Y[14]*Y[10]*X[10]*X[14]+2*Y[9]*Y[12]*X[12]*X[9]);
  Z[6][6]:=1/(thc[6]^2)*(+Y[6]^2*X[6]^2+Y[4]^2*X[4]^2-Y[16]^2*X[16]^2-Y[11]^2*X[11]^2-Y[15]^2*X[15]^2-Y[2]^2*X[2]^2+Y[10]^2*X[10]^2+Y[12]^2*X[12]^2+Y[5]^2*X[5]^2+Y[1]^2*X[1]^2-Y[8]^2*X[8]^2-Y[7]^2*X[7]^2-Y[13]^2*X[13]^2-Y[3]^2*X[3]^2+Y[14]^2*X[14]^2+Y[9]^2*X[9]^2);
  Z[6][7]:=1/(thc[6]*thc[7])*(2*Y[7]*Y[6]*X[6]*X[7]+2*Y[8]*Y[4]*X[4]*X[8]-2*Y[1]*Y[16]*X[16]*X[1]-2*Y[5]*Y[11]*X[11]*X[5]+2*Y[9]*Y[15]*X[15]*X[9]+2*Y[14]*Y[2]*X[2]*X[14]-2*Y[3]*Y[10]*X[10]*X[3]-2*Y[13]*Y[12]*X[12]*X[13]);
  Z[6][8]:=1/(thc[6]*thc[8])*(2*Y[8]*Y[6]*X[6]*X[8]+2*Y[7]*Y[4]*X[4]*X[7]-2*Y[5]*Y[16]*X[16]*X[5]-2*Y[1]*Y[11]*X[11]*X[1]-2*Y[14]*Y[15]*X[15]*X[14]-2*Y[9]*Y[2]*X[2]*X[9]+2*Y[13]*Y[10]*X[10]*X[13]+2*Y[3]*Y[12]*X[12]*X[3]);
  Z[6][9]:=1/(thc[6]*thc[9])*(2*Y[9]*Y[6]*X[6]*X[9]+2*Y[14]*Y[4]*X[4]*X[14]-2*Y[3]*Y[16]*X[16]*X[3]-2*Y[13]*Y[11]*X[11]*X[13]+2*Y[7]*Y[15]*X[15]*X[7]+2*Y[8]*Y[2]*X[2]*X[8]-2*Y[1]*Y[10]*X[10]*X[1]-2*Y[5]*Y[12]*X[12]*X[5]);
  Z[6][10]:=1/(thc[6]*thc[10])*(2*Y[10]*Y[6]*X[6]*X[10]+2*Y[12]*Y[4]*X[4]*X[12]-2*Y[15]*Y[16]*X[16]*X[15]-2*Y[2]*Y[11]*X[11]*X[2]-2*Y[14]*Y[5]*X[5]*X[14]-2*Y[9]*Y[1]*X[1]*X[9]+2*Y[13]*Y[8]*X[8]*X[13]+2*Y[3]*Y[7]*X[7]*X[3]);
  Z[6][11]:=1/(thc[5]*thc[7])*(2*Y[11]*Y[6]*X[5]*X[7]+2*Y[16]*Y[4]*X[1]*X[8]+2*Y[12]*Y[15]*X[13]*X[9]+2*Y[10]*Y[2]*X[3]*X[14]+2*Y[7]*Y[5]*X[6]*X[11]+2*Y[8]*Y[1]*X[4]*X[16]+2*Y[9]*Y[13]*X[15]*X[12]+2*Y[14]*Y[3]*X[2]*X[10]);
  Z[6][12]:=1/(thc[8]*thc[3])*(2*Y[12]*Y[6]*X[8]*X[3]+2*Y[10]*Y[4]*X[7]*X[13]+2*Y[2]*Y[16]*X[5]*X[9]+2*Y[15]*Y[11]*X[1]*X[14]+2*Y[9]*Y[5]*X[16]*X[2]+2*Y[14]*Y[1]*X[11]*X[15]+2*Y[3]*Y[8]*X[6]*X[12]+2*Y[13]*Y[7]*X[4]*X[10]);
  Z[6][13]:=1/(thc[10]*thc[8])*(2*Y[13]*Y[6]*X[10]*X[8]+2*Y[3]*Y[4]*X[12]*X[7]-2*Y[14]*Y[16]*X[15]*X[5]-2*Y[9]*Y[11]*X[2]*X[1]-2*Y[5]*Y[15]*X[16]*X[14]-2*Y[1]*Y[2]*X[11]*X[9]+2*Y[8]*Y[10]*X[6]*X[13]+2*Y[7]*Y[12]*X[4]*X[3]);
  Z[6][14]:=1/(thc[2]*thc[7])*(2*Y[14]*Y[6]*X[2]*X[7]+2*Y[9]*Y[4]*X[15]*X[8]-2*Y[13]*Y[16]*X[12]*X[1]-2*Y[3]*Y[11]*X[10]*X[5]+2*Y[8]*Y[15]*X[4]*X[9]+2*Y[7]*Y[2]*X[6]*X[14]-2*Y[5]*Y[10]*X[11]*X[3]-2*Y[1]*Y[12]*X[16]*X[13]);
  Z[6][15]:=1/(thc[1]*thc[3])*(2*Y[15]*Y[6]*X[1]*X[3]+2*Y[2]*Y[4]*X[5]*X[13]+2*Y[10]*Y[16]*X[7]*X[9]+2*Y[12]*Y[11]*X[8]*X[14]+2*Y[13]*Y[5]*X[4]*X[2]+2*Y[3]*Y[1]*X[6]*X[15]+2*Y[14]*Y[8]*X[11]*X[12]+2*Y[9]*Y[7]*X[16]*X[10]);
  Z[6][16]:=1/(thc[1]*thc[7])*(2*Y[16]*Y[6]*X[1]*X[7]+2*Y[11]*Y[4]*X[5]*X[8]-2*Y[10]*Y[15]*X[3]*X[9]-2*Y[12]*Y[2]*X[13]*X[14]+2*Y[8]*Y[5]*X[4]*X[11]+2*Y[7]*Y[1]*X[6]*X[16]-2*Y[14]*Y[13]*X[2]*X[12]-2*Y[9]*Y[3]*X[15]*X[10]);

  Z[7][1]:=1/(thc[7]*thc[1])*(2*Y[1]*Y[7]*X[7]*X[1]+2*Y[5]*Y[8]*X[8]*X[5]+2*Y[3]*Y[9]*X[9]*X[3]+2*Y[13]*Y[14]*X[14]*X[13]+2*Y[4]*Y[11]*X[11]*X[4]+2*Y[6]*Y[16]*X[16]*X[6]+2*Y[2]*Y[12]*X[12]*X[2]+2*Y[15]*Y[10]*X[10]*X[15]);
  Z[7][2]:=1/(thc[7]*thc[2])*(2*Y[2]*Y[7]*X[7]*X[2]+2*Y[15]*Y[8]*X[8]*X[15]+2*Y[12]*Y[1]*X[1]*X[12]+2*Y[10]*Y[5]*X[5]*X[10]+2*Y[4]*Y[9]*X[9]*X[4]+2*Y[6]*Y[14]*X[14]*X[6]+2*Y[11]*Y[3]*X[3]*X[11]+2*Y[16]*Y[13]*X[13]*X[16]);
  Z[7][3]:=1/(thc[7]*thc[3])*(2*Y[3]*Y[7]*X[7]*X[3]+2*Y[13]*Y[8]*X[8]*X[13]+2*Y[9]*Y[1]*X[1]*X[9]+2*Y[14]*Y[5]*X[5]*X[14]+2*Y[2]*Y[11]*X[11]*X[2]+2*Y[15]*Y[16]*X[16]*X[15]+2*Y[12]*Y[4]*X[4]*X[12]+2*Y[10]*Y[6]*X[6]*X[10]);
  Z[7][4]:=1/(thc[7]*thc[4])*(2*Y[4]*Y[7]*X[7]*X[4]+2*Y[6]*Y[8]*X[8]*X[6]+2*Y[11]*Y[1]*X[1]*X[11]+2*Y[16]*Y[5]*X[5]*X[16]+2*Y[2]*Y[9]*X[9]*X[2]+2*Y[15]*Y[14]*X[14]*X[15]+2*Y[12]*Y[3]*X[3]*X[12]+2*Y[10]*Y[13]*X[13]*X[10]);
  Z[7][5]:=1/(thc[7]*thc[5])*(2*Y[5]*Y[7]*X[7]*X[5]+2*Y[1]*Y[8]*X[8]*X[1]-2*Y[13]*Y[9]*X[9]*X[13]-2*Y[3]*Y[14]*X[14]*X[3]+2*Y[6]*Y[11]*X[11]*X[6]+2*Y[4]*Y[16]*X[16]*X[4]-2*Y[15]*Y[12]*X[12]*X[15]-2*Y[2]*Y[10]*X[10]*X[2]);
  Z[7][6]:=1/(thc[7]*thc[6])*(2*Y[6]*Y[7]*X[7]*X[6]+2*Y[4]*Y[8]*X[8]*X[4]+2*Y[16]*Y[1]*X[1]*X[16]+2*Y[11]*Y[5]*X[5]*X[11]-2*Y[15]*Y[9]*X[9]*X[15]-2*Y[2]*Y[14]*X[14]*X[2]-2*Y[10]*Y[3]*X[3]*X[10]-2*Y[12]*Y[13]*X[13]*X[12]);
  Z[7][7]:=1/(thc[7]^2)*(+Y[7]^2*X[7]^2+Y[8]^2*X[8]^2+Y[1]^2*X[1]^2+Y[5]^2*X[5]^2+Y[9]^2*X[9]^2+Y[14]^2*X[14]^2+Y[3]^2*X[3]^2+Y[13]^2*X[13]^2-Y[11]^2*X[11]^2-Y[16]^2*X[16]^2-Y[4]^2*X[4]^2-Y[6]^2*X[6]^2-Y[12]^2*X[12]^2-Y[10]^2*X[10]^2-Y[2]^2*X[2]^2-Y[15]^2*X[15]^2);
  Z[7][8]:=1/(thc[7]*thc[8])*(2*Y[8]*Y[7]*X[7]*X[8]+2*Y[5]*Y[1]*X[1]*X[5]-2*Y[14]*Y[9]*X[9]*X[14]-2*Y[13]*Y[3]*X[3]*X[13]-2*Y[16]*Y[11]*X[11]*X[16]-2*Y[6]*Y[4]*X[4]*X[6]+2*Y[10]*Y[12]*X[12]*X[10]+2*Y[15]*Y[2]*X[2]*X[15]);
  Z[7][9]:=1/(thc[7]*thc[9])*(2*Y[9]*Y[7]*X[7]*X[9]+2*Y[14]*Y[8]*X[8]*X[14]+2*Y[3]*Y[1]*X[1]*X[3]+2*Y[13]*Y[5]*X[5]*X[13]-2*Y[12]*Y[11]*X[11]*X[12]-2*Y[10]*Y[16]*X[16]*X[10]-2*Y[2]*Y[4]*X[4]*X[2]-2*Y[15]*Y[6]*X[6]*X[15]);
  Z[7][10]:=1/(thc[7]*thc[10])*(2*Y[10]*Y[7]*X[7]*X[10]+2*Y[12]*Y[8]*X[8]*X[12]+2*Y[15]*Y[1]*X[1]*X[15]+2*Y[2]*Y[5]*X[5]*X[2]-2*Y[16]*Y[9]*X[9]*X[16]-2*Y[11]*Y[14]*X[14]*X[11]-2*Y[6]*Y[3]*X[3]*X[6]-2*Y[4]*Y[13]*X[13]*X[4]);
  Z[7][11]:=1/(thc[5]*thc[6])*(2*Y[11]*Y[7]*X[5]*X[6]+2*Y[16]*Y[8]*X[1]*X[4]+2*Y[4]*Y[1]*X[8]*X[16]+2*Y[6]*Y[5]*X[7]*X[11]+2*Y[12]*Y[9]*X[13]*X[15]+2*Y[10]*Y[14]*X[3]*X[2]+2*Y[2]*Y[3]*X[14]*X[10]+2*Y[15]*Y[13]*X[9]*X[12]);
  Z[7][12]:=1/(thc[3]*thc[4])*(2*Y[12]*Y[7]*X[3]*X[4]+2*Y[10]*Y[8]*X[13]*X[6]+2*Y[2]*Y[1]*X[9]*X[11]+2*Y[15]*Y[5]*X[14]*X[16]+2*Y[11]*Y[9]*X[1]*X[2]+2*Y[16]*Y[14]*X[5]*X[15]+2*Y[4]*Y[3]*X[7]*X[12]+2*Y[6]*Y[13]*X[8]*X[10]);
  Z[7][13]:=1/(thc[8]*thc[3])*(2*Y[13]*Y[7]*X[8]*X[3]+2*Y[3]*Y[8]*X[7]*X[13]+2*Y[14]*Y[1]*X[5]*X[9]+2*Y[9]*Y[5]*X[1]*X[14]+2*Y[15]*Y[11]*X[16]*X[2]+2*Y[2]*Y[16]*X[11]*X[15]+2*Y[10]*Y[4]*X[6]*X[12]+2*Y[12]*Y[6]*X[4]*X[10]);
  Z[7][14]:=1/(thc[6]*thc[2])*(2*Y[14]*Y[7]*X[6]*X[2]+2*Y[9]*Y[8]*X[4]*X[15]-2*Y[13]*Y[1]*X[16]*X[12]-2*Y[3]*Y[5]*X[11]*X[10]-2*Y[10]*Y[11]*X[5]*X[3]-2*Y[12]*Y[16]*X[1]*X[13]+2*Y[15]*Y[4]*X[8]*X[9]+2*Y[2]*Y[6]*X[7]*X[14]);
  Z[7][15]:=1/(thc[8]*thc[2])*(2*Y[15]*Y[7]*X[8]*X[2]+2*Y[2]*Y[8]*X[7]*X[15]+2*Y[10]*Y[1]*X[5]*X[12]+2*Y[12]*Y[5]*X[1]*X[10]+2*Y[6]*Y[9]*X[14]*X[4]+2*Y[4]*Y[14]*X[9]*X[6]+2*Y[16]*Y[3]*X[13]*X[11]+2*Y[11]*Y[13]*X[3]*X[16]);
  Z[7][16]:=1/(thc[5]*thc[4])*(2*Y[16]*Y[7]*X[5]*X[4]+2*Y[11]*Y[8]*X[1]*X[6]+2*Y[6]*Y[1]*X[8]*X[11]+2*Y[4]*Y[5]*X[7]*X[16]-2*Y[10]*Y[9]*X[13]*X[2]-2*Y[12]*Y[14]*X[3]*X[15]-2*Y[15]*Y[3]*X[14]*X[12]-2*Y[2]*Y[13]*X[9]*X[10]);

  Z[8][1]:=1/(thc[8]*thc[1])*(2*Y[1]*Y[8]*X[8]*X[1]+2*Y[5]*Y[7]*X[7]*X[5]+2*Y[3]*Y[14]*X[14]*X[3]+2*Y[13]*Y[9]*X[9]*X[13]+2*Y[4]*Y[16]*X[16]*X[4]+2*Y[6]*Y[11]*X[11]*X[6]+2*Y[2]*Y[10]*X[10]*X[2]+2*Y[15]*Y[12]*X[12]*X[15]);
  Z[8][2]:=1/(thc[8]*thc[2])*(2*Y[2]*Y[8]*X[8]*X[2]+2*Y[15]*Y[7]*X[7]*X[15]+2*Y[12]*Y[5]*X[5]*X[12]+2*Y[10]*Y[1]*X[1]*X[10]+2*Y[4]*Y[14]*X[14]*X[4]+2*Y[6]*Y[9]*X[9]*X[6]+2*Y[11]*Y[13]*X[13]*X[11]+2*Y[16]*Y[3]*X[3]*X[16]);
  Z[8][3]:=1/(thc[8]*thc[3])*(2*Y[3]*Y[8]*X[8]*X[3]+2*Y[13]*Y[7]*X[7]*X[13]+2*Y[9]*Y[5]*X[5]*X[9]+2*Y[14]*Y[1]*X[1]*X[14]+2*Y[2]*Y[16]*X[16]*X[2]+2*Y[15]*Y[11]*X[11]*X[15]+2*Y[12]*Y[6]*X[6]*X[12]+2*Y[10]*Y[4]*X[4]*X[10]);
  Z[8][4]:=1/(thc[8]*thc[4])*(2*Y[4]*Y[8]*X[8]*X[4]+2*Y[6]*Y[7]*X[7]*X[6]+2*Y[11]*Y[5]*X[5]*X[11]+2*Y[16]*Y[1]*X[1]*X[16]+2*Y[2]*Y[14]*X[14]*X[2]+2*Y[15]*Y[9]*X[9]*X[15]+2*Y[12]*Y[13]*X[13]*X[12]+2*Y[10]*Y[3]*X[3]*X[10]);
  Z[8][5]:=1/(thc[8]*thc[5])*(2*Y[5]*Y[8]*X[8]*X[5]+2*Y[1]*Y[7]*X[7]*X[1]-2*Y[13]*Y[14]*X[14]*X[13]-2*Y[3]*Y[9]*X[9]*X[3]+2*Y[6]*Y[16]*X[16]*X[6]+2*Y[4]*Y[11]*X[11]*X[4]-2*Y[15]*Y[10]*X[10]*X[15]-2*Y[2]*Y[12]*X[12]*X[2]);
  Z[8][6]:=1/(thc[8]*thc[6])*(2*Y[6]*Y[8]*X[8]*X[6]+2*Y[4]*Y[7]*X[7]*X[4]+2*Y[16]*Y[5]*X[5]*X[16]+2*Y[11]*Y[1]*X[1]*X[11]-2*Y[15]*Y[14]*X[14]*X[15]-2*Y[2]*Y[9]*X[9]*X[2]-2*Y[10]*Y[13]*X[13]*X[10]-2*Y[12]*Y[3]*X[3]*X[12]);
  Z[8][7]:=1/(thc[8]*thc[7])*(2*Y[7]*Y[8]*X[8]*X[7]+2*Y[1]*Y[5]*X[5]*X[1]+2*Y[9]*Y[14]*X[14]*X[9]+2*Y[3]*Y[13]*X[13]*X[3]-2*Y[11]*Y[16]*X[16]*X[11]-2*Y[4]*Y[6]*X[6]*X[4]-2*Y[12]*Y[10]*X[10]*X[12]-2*Y[2]*Y[15]*X[15]*X[2]);
  Z[8][8]:=1/(thc[8]^2)*(+Y[8]^2*X[8]^2+Y[7]^2*X[7]^2+Y[5]^2*X[5]^2+Y[1]^2*X[1]^2-Y[14]^2*X[14]^2-Y[9]^2*X[9]^2-Y[13]^2*X[13]^2-Y[3]^2*X[3]^2-Y[16]^2*X[16]^2-Y[11]^2*X[11]^2-Y[6]^2*X[6]^2-Y[4]^2*X[4]^2+Y[10]^2*X[10]^2+Y[12]^2*X[12]^2+Y[15]^2*X[15]^2+Y[2]^2*X[2]^2);
  Z[8][9]:=1/(thc[8]*thc[9])*(2*Y[9]*Y[8]*X[8]*X[9]+2*Y[14]*Y[7]*X[7]*X[14]+2*Y[3]*Y[5]*X[5]*X[3]+2*Y[13]*Y[1]*X[1]*X[13]-2*Y[12]*Y[16]*X[16]*X[12]-2*Y[10]*Y[11]*X[11]*X[10]-2*Y[2]*Y[6]*X[6]*X[2]-2*Y[15]*Y[4]*X[4]*X[15]);
  Z[8][10]:=1/(thc[8]*thc[10])*(2*Y[10]*Y[8]*X[8]*X[10]+2*Y[12]*Y[7]*X[7]*X[12]+2*Y[15]*Y[5]*X[5]*X[15]+2*Y[2]*Y[1]*X[1]*X[2]-2*Y[16]*Y[14]*X[14]*X[16]-2*Y[11]*Y[9]*X[9]*X[11]-2*Y[6]*Y[13]*X[13]*X[6]-2*Y[4]*Y[3]*X[3]*X[4]);
  Z[8][11]:=1/(thc[1]*thc[6])*(2*Y[11]*Y[8]*X[1]*X[6]+2*Y[16]*Y[7]*X[5]*X[4]+2*Y[4]*Y[5]*X[7]*X[16]+2*Y[6]*Y[1]*X[8]*X[11]+2*Y[12]*Y[14]*X[3]*X[15]+2*Y[10]*Y[9]*X[13]*X[2]+2*Y[2]*Y[13]*X[9]*X[10]+2*Y[15]*Y[3]*X[14]*X[12]);
  Z[8][12]:=1/(thc[3]*thc[6])*(2*Y[12]*Y[8]*X[3]*X[6]+2*Y[10]*Y[7]*X[13]*X[4]+2*Y[2]*Y[5]*X[9]*X[16]+2*Y[15]*Y[1]*X[14]*X[11]+2*Y[11]*Y[14]*X[1]*X[15]+2*Y[16]*Y[9]*X[5]*X[2]+2*Y[4]*Y[13]*X[7]*X[10]+2*Y[6]*Y[3]*X[8]*X[12]);
  Z[8][13]:=1/(thc[1]*thc[9])*(2*Y[13]*Y[8]*X[1]*X[9]+2*Y[3]*Y[7]*X[5]*X[14]+2*Y[14]*Y[5]*X[7]*X[3]+2*Y[9]*Y[1]*X[8]*X[13]+2*Y[15]*Y[16]*X[4]*X[12]+2*Y[2]*Y[11]*X[6]*X[10]+2*Y[10]*Y[6]*X[11]*X[2]+2*Y[12]*Y[4]*X[16]*X[15]);
  Z[8][14]:=1/(thc[1]*thc[3])*(2*Y[14]*Y[8]*X[1]*X[3]+2*Y[9]*Y[7]*X[5]*X[13]+2*Y[13]*Y[5]*X[7]*X[9]+2*Y[3]*Y[1]*X[8]*X[14]-2*Y[10]*Y[16]*X[4]*X[2]-2*Y[12]*Y[11]*X[6]*X[15]-2*Y[15]*Y[6]*X[11]*X[12]-2*Y[2]*Y[4]*X[16]*X[10]);
  Z[8][15]:=1/(thc[9]*thc[4])*(2*Y[15]*Y[8]*X[9]*X[4]+2*Y[2]*Y[7]*X[14]*X[6]+2*Y[10]*Y[5]*X[3]*X[11]+2*Y[12]*Y[1]*X[13]*X[16]+2*Y[6]*Y[14]*X[7]*X[2]+2*Y[4]*Y[9]*X[8]*X[15]+2*Y[16]*Y[13]*X[1]*X[12]+2*Y[11]*Y[3]*X[5]*X[10]);
  Z[8][16]:=1/(thc[5]*thc[6])*(2*Y[16]*Y[8]*X[5]*X[6]+2*Y[11]*Y[7]*X[1]*X[4]+2*Y[6]*Y[5]*X[8]*X[16]+2*Y[4]*Y[1]*X[7]*X[11]-2*Y[10]*Y[14]*X[13]*X[15]-2*Y[12]*Y[9]*X[3]*X[2]-2*Y[15]*Y[13]*X[14]*X[10]-2*Y[2]*Y[3]*X[9]*X[12]);

  Z[9][1]:=1/(thc[9]*thc[1])*(2*Y[1]*Y[9]*X[9]*X[1]-2*Y[5]*Y[14]*X[14]*X[5]+2*Y[7]*Y[3]*X[3]*X[7]-2*Y[8]*Y[13]*X[13]*X[8]+2*Y[4]*Y[12]*X[12]*X[4]-2*Y[6]*Y[10]*X[10]*X[6]+2*Y[11]*Y[2]*X[2]*X[11]-2*Y[16]*Y[15]*X[15]*X[16]);
  Z[9][2]:=1/(thc[9]*thc[2])*(2*Y[2]*Y[9]*X[9]*X[2]-2*Y[15]*Y[14]*X[14]*X[15]+2*Y[12]*Y[3]*X[3]*X[12]-2*Y[10]*Y[13]*X[13]*X[10]+2*Y[4]*Y[7]*X[7]*X[4]-2*Y[6]*Y[8]*X[8]*X[6]+2*Y[11]*Y[1]*X[1]*X[11]-2*Y[16]*Y[5]*X[5]*X[16]);
  Z[9][3]:=1/(thc[9]*thc[3])*(2*Y[3]*Y[9]*X[9]*X[3]-2*Y[13]*Y[14]*X[14]*X[13]+2*Y[1]*Y[7]*X[7]*X[1]-2*Y[5]*Y[8]*X[8]*X[5]+2*Y[2]*Y[12]*X[12]*X[2]-2*Y[15]*Y[10]*X[10]*X[15]+2*Y[4]*Y[11]*X[11]*X[4]-2*Y[6]*Y[16]*X[16]*X[6]);
  Z[9][4]:=1/(thc[9]*thc[4])*(2*Y[4]*Y[9]*X[9]*X[4]-2*Y[6]*Y[14]*X[14]*X[6]+2*Y[11]*Y[3]*X[3]*X[11]-2*Y[16]*Y[13]*X[13]*X[16]+2*Y[2]*Y[7]*X[7]*X[2]-2*Y[15]*Y[8]*X[8]*X[15]+2*Y[12]*Y[1]*X[1]*X[12]-2*Y[10]*Y[5]*X[5]*X[10]);
  Z[9][5]:=1/(thc[9]*thc[5])*(2*Y[5]*Y[9]*X[9]*X[5]-2*Y[1]*Y[14]*X[14]*X[1]+2*Y[8]*Y[3]*X[3]*X[8]-2*Y[7]*Y[13]*X[13]*X[7]+2*Y[6]*Y[12]*X[12]*X[6]-2*Y[4]*Y[10]*X[10]*X[4]+2*Y[16]*Y[2]*X[2]*X[16]-2*Y[11]*Y[15]*X[15]*X[11]);
  Z[9][6]:=1/(thc[9]*thc[6])*(2*Y[6]*Y[9]*X[9]*X[6]-2*Y[4]*Y[14]*X[14]*X[4]+2*Y[16]*Y[3]*X[3]*X[16]-2*Y[11]*Y[13]*X[13]*X[11]-2*Y[15]*Y[7]*X[7]*X[15]+2*Y[2]*Y[8]*X[8]*X[2]-2*Y[10]*Y[1]*X[1]*X[10]+2*Y[12]*Y[5]*X[5]*X[12]);
  Z[9][7]:=1/(thc[9]*thc[7])*(2*Y[7]*Y[9]*X[9]*X[7]-2*Y[8]*Y[14]*X[14]*X[8]+2*Y[1]*Y[3]*X[3]*X[1]-2*Y[5]*Y[13]*X[13]*X[5]-2*Y[11]*Y[12]*X[12]*X[11]+2*Y[16]*Y[10]*X[10]*X[16]-2*Y[4]*Y[2]*X[2]*X[4]+2*Y[6]*Y[15]*X[15]*X[6]);
  Z[9][8]:=1/(thc[9]*thc[8])*(2*Y[8]*Y[9]*X[9]*X[8]-2*Y[7]*Y[14]*X[14]*X[7]+2*Y[5]*Y[3]*X[3]*X[5]-2*Y[1]*Y[13]*X[13]*X[1]-2*Y[16]*Y[12]*X[12]*X[16]+2*Y[11]*Y[10]*X[10]*X[11]-2*Y[6]*Y[2]*X[2]*X[6]+2*Y[4]*Y[15]*X[15]*X[4]);
  Z[9][9]:=1/(thc[9]^2)*(+Y[9]^2*X[9]^2-Y[14]^2*X[14]^2+Y[3]^2*X[3]^2-Y[13]^2*X[13]^2+Y[7]^2*X[7]^2-Y[8]^2*X[8]^2+Y[1]^2*X[1]^2-Y[5]^2*X[5]^2-Y[12]^2*X[12]^2+Y[10]^2*X[10]^2-Y[2]^2*X[2]^2+Y[15]^2*X[15]^2-Y[11]^2*X[11]^2+Y[16]^2*X[16]^2-Y[4]^2*X[4]^2+Y[6]^2*X[6]^2);
  Z[9][10]:=1/(thc[9]*thc[10])*(2*Y[10]*Y[9]*X[9]*X[10]-2*Y[12]*Y[14]*X[14]*X[12]+2*Y[15]*Y[3]*X[3]*X[15]-2*Y[2]*Y[13]*X[13]*X[2]-2*Y[16]*Y[7]*X[7]*X[16]+2*Y[11]*Y[8]*X[8]*X[11]-2*Y[6]*Y[1]*X[1]*X[6]+2*Y[4]*Y[5]*X[5]*X[4]);
  Z[9][11]:=1/(thc[8]*thc[10])*(2*Y[11]*Y[9]*X[8]*X[10]-2*Y[16]*Y[14]*X[7]*X[12]+2*Y[4]*Y[3]*X[5]*X[15]-2*Y[6]*Y[13]*X[1]*X[2]-2*Y[12]*Y[7]*X[14]*X[16]+2*Y[10]*Y[8]*X[9]*X[11]-2*Y[2]*Y[1]*X[13]*X[6]+2*Y[15]*Y[5]*X[3]*X[4]);
  Z[9][12]:=1/(thc[1]*thc[4])*(2*Y[12]*Y[9]*X[1]*X[4]+2*Y[10]*Y[14]*X[5]*X[6]+2*Y[2]*Y[3]*X[7]*X[11]+2*Y[15]*Y[13]*X[8]*X[16]+2*Y[11]*Y[7]*X[3]*X[2]+2*Y[16]*Y[8]*X[13]*X[15]+2*Y[4]*Y[1]*X[9]*X[12]+2*Y[6]*Y[5]*X[14]*X[10]);
  Z[9][13]:=1/(thc[7]*thc[5])*(2*Y[13]*Y[9]*X[7]*X[5]+2*Y[3]*Y[14]*X[8]*X[1]+2*Y[5]*Y[7]*X[9]*X[13]+2*Y[1]*Y[8]*X[14]*X[3]+2*Y[15]*Y[12]*X[11]*X[6]+2*Y[2]*Y[10]*X[16]*X[4]+2*Y[6]*Y[11]*X[12]*X[15]+2*Y[4]*Y[16]*X[10]*X[2]);
  Z[9][14]:=1/(thc[7]*thc[8])*(2*Y[14]*Y[9]*X[7]*X[8]+2*Y[13]*Y[3]*X[1]*X[5]+2*Y[8]*Y[7]*X[9]*X[14]+2*Y[5]*Y[1]*X[3]*X[13]-2*Y[10]*Y[12]*X[11]*X[16]-2*Y[15]*Y[2]*X[4]*X[6]-2*Y[16]*Y[11]*X[12]*X[10]-2*Y[6]*Y[4]*X[2]*X[15]);
  Z[9][15]:=1/(thc[3]*thc[10])*(2*Y[15]*Y[9]*X[3]*X[10]-2*Y[2]*Y[14]*X[13]*X[12]+2*Y[10]*Y[3]*X[9]*X[15]-2*Y[12]*Y[13]*X[14]*X[2]-2*Y[6]*Y[7]*X[1]*X[16]+2*Y[4]*Y[8]*X[5]*X[11]-2*Y[16]*Y[1]*X[7]*X[6]+2*Y[11]*Y[5]*X[8]*X[4]);
  Z[9][16]:=1/(thc[7]*thc[10])*(2*Y[16]*Y[9]*X[7]*X[10]-2*Y[11]*Y[14]*X[8]*X[12]+2*Y[6]*Y[3]*X[1]*X[15]-2*Y[4]*Y[13]*X[5]*X[2]+2*Y[10]*Y[7]*X[9]*X[16]-2*Y[12]*Y[8]*X[14]*X[11]+2*Y[15]*Y[1]*X[3]*X[6]-2*Y[2]*Y[5]*X[13]*X[4]);

  Z[10][1]:=1/(thc[10]*thc[1])*(2*Y[1]*Y[10]*X[10]*X[1]-2*Y[5]*Y[12]*X[12]*X[5]-2*Y[7]*Y[15]*X[15]*X[7]+2*Y[8]*Y[2]*X[2]*X[8]+2*Y[3]*Y[16]*X[16]*X[3]-2*Y[13]*Y[11]*X[11]*X[13]-2*Y[9]*Y[6]*X[6]*X[9]+2*Y[14]*Y[4]*X[4]*X[14]);
  Z[10][2]:=1/(thc[10]*thc[2])*(2*Y[2]*Y[10]*X[10]*X[2]-2*Y[15]*Y[12]*X[12]*X[15]+2*Y[4]*Y[16]*X[16]*X[4]-2*Y[6]*Y[11]*X[11]*X[6]+2*Y[3]*Y[14]*X[14]*X[3]-2*Y[13]*Y[9]*X[9]*X[13]+2*Y[1]*Y[8]*X[8]*X[1]-2*Y[5]*Y[7]*X[7]*X[5]);
  Z[10][3]:=1/(thc[10]*thc[3])*(2*Y[3]*Y[10]*X[10]*X[3]-2*Y[13]*Y[12]*X[12]*X[13]-2*Y[9]*Y[15]*X[15]*X[9]+2*Y[14]*Y[2]*X[2]*X[14]+2*Y[1]*Y[16]*X[16]*X[1]-2*Y[5]*Y[11]*X[11]*X[5]-2*Y[7]*Y[6]*X[6]*X[7]+2*Y[8]*Y[4]*X[4]*X[8]);
  Z[10][4]:=1/(thc[10]*thc[4])*(2*Y[4]*Y[10]*X[10]*X[4]-2*Y[6]*Y[12]*X[12]*X[6]-2*Y[11]*Y[15]*X[15]*X[11]+2*Y[16]*Y[2]*X[2]*X[16]+2*Y[1]*Y[14]*X[14]*X[1]-2*Y[5]*Y[9]*X[9]*X[5]-2*Y[7]*Y[13]*X[13]*X[7]+2*Y[8]*Y[3]*X[3]*X[8]);
  Z[10][5]:=1/(thc[10]*thc[5])*(2*Y[5]*Y[10]*X[10]*X[5]-2*Y[1]*Y[12]*X[12]*X[1]-2*Y[8]*Y[15]*X[15]*X[8]+2*Y[7]*Y[2]*X[2]*X[7]-2*Y[13]*Y[16]*X[16]*X[13]+2*Y[3]*Y[11]*X[11]*X[3]+2*Y[14]*Y[6]*X[6]*X[14]-2*Y[9]*Y[4]*X[4]*X[9]);
  Z[10][6]:=1/(thc[10]*thc[6])*(2*Y[6]*Y[10]*X[10]*X[6]-2*Y[4]*Y[12]*X[12]*X[4]-2*Y[16]*Y[15]*X[15]*X[16]+2*Y[11]*Y[2]*X[2]*X[11]+2*Y[5]*Y[14]*X[14]*X[5]-2*Y[1]*Y[9]*X[9]*X[1]-2*Y[8]*Y[13]*X[13]*X[8]+2*Y[7]*Y[3]*X[3]*X[7]);
  Z[10][7]:=1/(thc[10]*thc[7])*(2*Y[7]*Y[10]*X[10]*X[7]-2*Y[8]*Y[12]*X[12]*X[8]-2*Y[1]*Y[15]*X[15]*X[1]+2*Y[5]*Y[2]*X[2]*X[5]+2*Y[9]*Y[16]*X[16]*X[9]-2*Y[14]*Y[11]*X[11]*X[14]-2*Y[3]*Y[6]*X[6]*X[3]+2*Y[13]*Y[4]*X[4]*X[13]);
  Z[10][8]:=1/(thc[10]*thc[8])*(2*Y[8]*Y[10]*X[10]*X[8]-2*Y[7]*Y[12]*X[12]*X[7]-2*Y[5]*Y[15]*X[15]*X[5]+2*Y[1]*Y[2]*X[2]*X[1]-2*Y[14]*Y[16]*X[16]*X[14]+2*Y[9]*Y[11]*X[11]*X[9]+2*Y[13]*Y[6]*X[6]*X[13]-2*Y[3]*Y[4]*X[4]*X[3]);
  Z[10][9]:=1/(thc[10]*thc[9])*(2*Y[9]*Y[10]*X[10]*X[9]-2*Y[14]*Y[12]*X[12]*X[14]-2*Y[3]*Y[15]*X[15]*X[3]+2*Y[13]*Y[2]*X[2]*X[13]+2*Y[7]*Y[16]*X[16]*X[7]-2*Y[8]*Y[11]*X[11]*X[8]-2*Y[1]*Y[6]*X[6]*X[1]+2*Y[5]*Y[4]*X[4]*X[5]);
  Z[10][10]:=1/(thc[10]^2)*(+Y[10]^2*X[10]^2-Y[12]^2*X[12]^2-Y[15]^2*X[15]^2+Y[2]^2*X[2]^2-Y[16]^2*X[16]^2+Y[11]^2*X[11]^2+Y[6]^2*X[6]^2-Y[4]^2*X[4]^2-Y[14]^2*X[14]^2+Y[9]^2*X[9]^2+Y[13]^2*X[13]^2-Y[3]^2*X[3]^2+Y[8]^2*X[8]^2-Y[7]^2*X[7]^2-Y[5]^2*X[5]^2+Y[1]^2*X[1]^2);
  Z[10][11]:=1/(thc[3]*thc[5])*(2*Y[11]*Y[10]*X[3]*X[5]-2*Y[16]*Y[12]*X[13]*X[1]+2*Y[4]*Y[15]*X[9]*X[8]-2*Y[6]*Y[2]*X[14]*X[7]-2*Y[7]*Y[14]*X[2]*X[6]+2*Y[8]*Y[9]*X[15]*X[4]-2*Y[1]*Y[13]*X[12]*X[16]+2*Y[5]*Y[3]*X[10]*X[11]);
  Z[10][12]:=1/(thc[6]*thc[4])*(2*Y[12]*Y[10]*X[6]*X[4]-2*Y[2]*Y[15]*X[16]*X[11]-2*Y[11]*Y[16]*X[15]*X[2]+2*Y[4]*Y[6]*X[10]*X[12]-2*Y[9]*Y[14]*X[5]*X[1]+2*Y[3]*Y[13]*X[8]*X[7]+2*Y[7]*Y[8]*X[13]*X[3]-2*Y[1]*Y[5]*X[14]*X[9]);
  Z[10][13]:=1/(thc[4]*thc[7])*(2*Y[13]*Y[10]*X[4]*X[7]+2*Y[3]*Y[12]*X[6]*X[8]-2*Y[14]*Y[15]*X[11]*X[1]-2*Y[9]*Y[2]*X[16]*X[5]-2*Y[5]*Y[16]*X[2]*X[9]-2*Y[1]*Y[11]*X[15]*X[14]+2*Y[8]*Y[6]*X[12]*X[3]+2*Y[7]*Y[4]*X[10]*X[13]);
  Z[10][14]:=1/(thc[2]*thc[3])*(2*Y[14]*Y[10]*X[2]*X[3]-2*Y[9]*Y[12]*X[15]*X[13]-2*Y[13]*Y[15]*X[12]*X[9]+2*Y[3]*Y[2]*X[10]*X[14]-2*Y[8]*Y[16]*X[4]*X[1]+2*Y[7]*Y[11]*X[6]*X[5]+2*Y[5]*Y[6]*X[11]*X[7]-2*Y[1]*Y[4]*X[16]*X[8]);
  Z[10][15]:=1/(thc[8]*thc[5])*(2*Y[15]*Y[10]*X[8]*X[5]+2*Y[2]*Y[12]*X[7]*X[1]-2*Y[6]*Y[16]*X[14]*X[13]-2*Y[4]*Y[11]*X[9]*X[3]-2*Y[13]*Y[14]*X[16]*X[6]-2*Y[3]*Y[9]*X[11]*X[4]+2*Y[5]*Y[8]*X[10]*X[15]+2*Y[1]*Y[7]*X[12]*X[2]);
  Z[10][16]:=1/(thc[3]*thc[1])*(2*Y[16]*Y[10]*X[3]*X[1]-2*Y[11]*Y[12]*X[13]*X[5]+2*Y[6]*Y[15]*X[9]*X[7]-2*Y[4]*Y[2]*X[14]*X[8]-2*Y[8]*Y[14]*X[2]*X[4]+2*Y[7]*Y[9]*X[15]*X[6]-2*Y[5]*Y[13]*X[12]*X[11]+2*Y[1]*Y[3]*X[10]*X[16]);

  Z[11][1]:=1/(thc[6]*thc[8])*(-2*Y[1]*Y[11]*X[6]*X[8]-2*Y[5]*Y[16]*X[4]*X[7]+2*Y[7]*Y[4]*X[16]*X[5]+2*Y[8]*Y[6]*X[11]*X[1]-2*Y[3]*Y[12]*X[15]*X[14]-2*Y[13]*Y[10]*X[2]*X[9]+2*Y[9]*Y[2]*X[10]*X[13]+2*Y[14]*Y[15]*X[12]*X[3]);
  Z[11][2]:=1/(thc[1]*thc[9])*(-2*Y[2]*Y[11]*X[1]*X[9]-2*Y[15]*Y[16]*X[5]*X[14]-2*Y[12]*Y[4]*X[7]*X[3]-2*Y[10]*Y[6]*X[8]*X[13]+2*Y[3]*Y[7]*X[4]*X[12]+2*Y[13]*Y[8]*X[6]*X[10]+2*Y[9]*Y[1]*X[11]*X[2]+2*Y[14]*Y[5]*X[16]*X[15]);
  Z[11][3]:=1/(thc[2]*thc[7])*(-2*Y[3]*Y[11]*X[2]*X[7]-2*Y[13]*Y[16]*X[15]*X[8]+2*Y[9]*Y[4]*X[12]*X[1]+2*Y[14]*Y[6]*X[10]*X[5]-2*Y[1]*Y[12]*X[4]*X[9]-2*Y[5]*Y[10]*X[6]*X[14]+2*Y[7]*Y[2]*X[11]*X[3]+2*Y[8]*Y[15]*X[16]*X[13]);
  Z[11][4]:=1/(thc[7]*thc[1])*(-2*Y[4]*Y[11]*X[7]*X[1]-2*Y[6]*Y[16]*X[8]*X[5]-2*Y[2]*Y[12]*X[9]*X[3]-2*Y[15]*Y[10]*X[14]*X[13]+2*Y[1]*Y[7]*X[11]*X[4]+2*Y[5]*Y[8]*X[16]*X[6]+2*Y[3]*Y[9]*X[12]*X[2]+2*Y[13]*Y[14]*X[10]*X[15]);
  Z[11][5]:=1/(thc[4]*thc[8])*(-2*Y[5]*Y[11]*X[4]*X[8]-2*Y[1]*Y[16]*X[6]*X[7]+2*Y[8]*Y[4]*X[11]*X[5]+2*Y[7]*Y[6]*X[16]*X[1]+2*Y[13]*Y[12]*X[2]*X[14]+2*Y[3]*Y[10]*X[15]*X[9]-2*Y[14]*Y[2]*X[12]*X[13]-2*Y[9]*Y[15]*X[10]*X[3]);
  Z[11][6]:=1/(thc[7]*thc[5])*(-2*Y[6]*Y[11]*X[7]*X[5]-2*Y[4]*Y[16]*X[8]*X[1]+2*Y[15]*Y[12]*X[9]*X[13]+2*Y[2]*Y[10]*X[14]*X[3]+2*Y[5]*Y[7]*X[11]*X[6]+2*Y[1]*Y[8]*X[16]*X[4]-2*Y[13]*Y[9]*X[12]*X[15]-2*Y[3]*Y[14]*X[10]*X[2]);
  Z[11][7]:=1/(thc[2]*thc[3])*(-2*Y[7]*Y[11]*X[2]*X[3]+2*Y[8]*Y[16]*X[15]*X[13]+2*Y[1]*Y[4]*X[12]*X[9]-2*Y[5]*Y[6]*X[10]*X[14]-2*Y[9]*Y[12]*X[4]*X[1]+2*Y[14]*Y[10]*X[6]*X[5]+2*Y[3]*Y[2]*X[11]*X[7]-2*Y[13]*Y[15]*X[16]*X[8]);
  Z[11][8]:=1/(thc[6]*thc[1])*(-2*Y[8]*Y[11]*X[6]*X[1]-2*Y[7]*Y[16]*X[4]*X[5]+2*Y[5]*Y[4]*X[16]*X[7]+2*Y[1]*Y[6]*X[11]*X[8]+2*Y[14]*Y[12]*X[15]*X[3]+2*Y[9]*Y[10]*X[2]*X[13]-2*Y[13]*Y[2]*X[10]*X[9]-2*Y[3]*Y[15]*X[12]*X[14]);
  Z[11][9]:=1/(thc[4]*thc[3])*(-2*Y[9]*Y[11]*X[4]*X[3]-2*Y[14]*Y[16]*X[6]*X[13]+2*Y[3]*Y[4]*X[11]*X[9]+2*Y[13]*Y[6]*X[16]*X[14]-2*Y[7]*Y[12]*X[2]*X[1]-2*Y[8]*Y[10]*X[15]*X[5]+2*Y[1]*Y[2]*X[12]*X[7]+2*Y[5]*Y[15]*X[10]*X[8]);
  Z[11][10]:=1/(thc[6]*thc[2])*(-2*Y[10]*Y[11]*X[6]*X[2]-2*Y[12]*Y[16]*X[4]*X[15]+2*Y[15]*Y[4]*X[16]*X[12]+2*Y[2]*Y[6]*X[11]*X[10]+2*Y[14]*Y[7]*X[5]*X[3]+2*Y[9]*Y[8]*X[1]*X[13]-2*Y[13]*Y[1]*X[8]*X[9]-2*Y[3]*Y[5]*X[7]*X[14]);
  Z[11][11]:=1/(thc[1]^2)*(-Y[11]^2*X[1]^2-Y[16]^2*X[5]^2-Y[4]^2*X[7]^2-Y[6]^2*X[8]^2-Y[12]^2*X[3]^2-Y[10]^2*X[13]^2-Y[2]^2*X[9]^2-Y[15]^2*X[14]^2+Y[7]^2*X[4]^2+Y[8]^2*X[6]^2+Y[1]^2*X[11]^2+Y[5]^2*X[16]^2+Y[9]^2*X[2]^2+Y[14]^2*X[15]^2+Y[3]^2*X[12]^2+Y[13]^2*X[10]^2);
  Z[11][12]:=1/(thc[7]*thc[9])*(-2*Y[12]*Y[11]*X[7]*X[9]-2*Y[10]*Y[16]*X[8]*X[14]-2*Y[2]*Y[4]*X[1]*X[3]-2*Y[15]*Y[6]*X[5]*X[13]+2*Y[9]*Y[7]*X[11]*X[12]+2*Y[14]*Y[8]*X[16]*X[10]+2*Y[3]*Y[1]*X[4]*X[2]+2*Y[13]*Y[5]*X[6]*X[15]);
  Z[11][13]:=1/(thc[2]*thc[8])*(-2*Y[13]*Y[11]*X[2]*X[8]-2*Y[3]*Y[16]*X[15]*X[7]+2*Y[14]*Y[4]*X[12]*X[5]+2*Y[9]*Y[6]*X[10]*X[1]-2*Y[5]*Y[12]*X[4]*X[14]-2*Y[1]*Y[10]*X[6]*X[9]+2*Y[8]*Y[2]*X[11]*X[13]+2*Y[7]*Y[15]*X[16]*X[3]);
  Z[11][14]:=-1/(thc[6]*thc[3])*(2*Y[14]*Y[11]*X[6]*X[3]+2*Y[9]*Y[16]*X[4]*X[13]-2*Y[13]*Y[4]*X[16]*X[9]-2*Y[3]*Y[6]*X[11]*X[14]+2*Y[8]*Y[12]*X[15]*X[1]+2*Y[7]*Y[10]*X[2]*X[5]-2*Y[5]*Y[2]*X[10]*X[7]-2*Y[1]*Y[15]*X[12]*X[8]);
  Z[11][15]:=1/(thc[5]*thc[9])*(-2*Y[15]*Y[11]*X[5]*X[9]-2*Y[2]*Y[16]*X[1]*X[14]-2*Y[10]*Y[4]*X[8]*X[3]-2*Y[12]*Y[6]*X[7]*X[13]+2*Y[13]*Y[7]*X[6]*X[12]+2*Y[3]*Y[8]*X[4]*X[10]+2*Y[14]*Y[1]*X[16]*X[2]+2*Y[9]*Y[5]*X[11]*X[15]);
  Z[11][16]:=1/(thc[1]*thc[5])*(-2*Y[16]*Y[11]*X[1]*X[5]-2*Y[6]*Y[4]*X[7]*X[8]+2*Y[10]*Y[12]*X[3]*X[13]+2*Y[15]*Y[2]*X[9]*X[14]+2*Y[8]*Y[7]*X[4]*X[6]+2*Y[5]*Y[1]*X[11]*X[16]-2*Y[14]*Y[9]*X[2]*X[15]-2*Y[13]*Y[3]*X[12]*X[10]);

  Z[12][1]:=1/(thc[10]*thc[5])*(-2*Y[1]*Y[12]*X[10]*X[5]+2*Y[5]*Y[10]*X[12]*X[1]+2*Y[7]*Y[2]*X[15]*X[8]-2*Y[8]*Y[15]*X[2]*X[7]+2*Y[3]*Y[11]*X[16]*X[13]-2*Y[13]*Y[16]*X[11]*X[3]-2*Y[9]*Y[4]*X[6]*X[14]+2*Y[14]*Y[6]*X[4]*X[9]);
  Z[12][2]:=1/(thc[9]*thc[3])*(-2*Y[2]*Y[12]*X[9]*X[3]+2*Y[15]*Y[10]*X[14]*X[13]-2*Y[4]*Y[11]*X[7]*X[1]+2*Y[6]*Y[16]*X[8]*X[5]+2*Y[3]*Y[9]*X[12]*X[2]-2*Y[13]*Y[14]*X[10]*X[15]+2*Y[1]*Y[7]*X[11]*X[4]-2*Y[5]*Y[8]*X[16]*X[6]);
  Z[12][3]:=1/(thc[4]*thc[7])*(-2*Y[3]*Y[12]*X[4]*X[7]-2*Y[13]*Y[10]*X[6]*X[8]+2*Y[9]*Y[2]*X[11]*X[1]+2*Y[14]*Y[15]*X[16]*X[5]-2*Y[1]*Y[11]*X[2]*X[9]-2*Y[5]*Y[16]*X[15]*X[14]+2*Y[7]*Y[4]*X[12]*X[3]+2*Y[8]*Y[6]*X[10]*X[13]);
  Z[12][4]:=1/(thc[3]*thc[7])*(-2*Y[4]*Y[12]*X[3]*X[7]+2*Y[6]*Y[10]*X[13]*X[8]-2*Y[11]*Y[2]*X[9]*X[1]+2*Y[16]*Y[15]*X[14]*X[5]+2*Y[1]*Y[9]*X[2]*X[11]-2*Y[5]*Y[14]*X[15]*X[16]+2*Y[7]*Y[3]*X[12]*X[4]-2*Y[8]*Y[13]*X[10]*X[6]);
  Z[12][5]:=1/(thc[2]*thc[8])*(-2*Y[5]*Y[12]*X[2]*X[8]+2*Y[1]*Y[10]*X[15]*X[7]+2*Y[8]*Y[2]*X[12]*X[5]-2*Y[7]*Y[15]*X[10]*X[1]+2*Y[13]*Y[11]*X[4]*X[14]-2*Y[3]*Y[16]*X[6]*X[9]-2*Y[14]*Y[4]*X[11]*X[13]+2*Y[9]*Y[6]*X[16]*X[3]);
  Z[12][6]:=1/(thc[3]*thc[8])*(-2*Y[6]*Y[12]*X[3]*X[8]+2*Y[4]*Y[10]*X[13]*X[7]-2*Y[16]*Y[2]*X[9]*X[5]+2*Y[11]*Y[15]*X[14]*X[1]+2*Y[5]*Y[9]*X[2]*X[16]-2*Y[1]*Y[14]*X[15]*X[11]+2*Y[8]*Y[3]*X[12]*X[6]-2*Y[7]*Y[13]*X[10]*X[4]);
  Z[12][7]:=1/(thc[10]*thc[8])*(-2*Y[7]*Y[12]*X[10]*X[8]+2*Y[8]*Y[10]*X[12]*X[7]+2*Y[1]*Y[2]*X[15]*X[5]-2*Y[5]*Y[15]*X[2]*X[1]+2*Y[9]*Y[11]*X[16]*X[14]-2*Y[14]*Y[16]*X[11]*X[9]-2*Y[3]*Y[4]*X[6]*X[13]+2*Y[13]*Y[6]*X[4]*X[3]);
  Z[12][8]:=1/(thc[10]*thc[7])*(-2*Y[8]*Y[12]*X[10]*X[7]+2*Y[7]*Y[10]*X[12]*X[8]+2*Y[5]*Y[2]*X[15]*X[1]-2*Y[1]*Y[15]*X[2]*X[5]-2*Y[14]*Y[11]*X[16]*X[9]+2*Y[9]*Y[16]*X[11]*X[14]+2*Y[13]*Y[4]*X[6]*X[3]-2*Y[3]*Y[6]*X[4]*X[13]);
  Z[12][9]:=1/(thc[4]*thc[1])*(-2*Y[9]*Y[12]*X[4]*X[1]-2*Y[14]*Y[10]*X[6]*X[5]+2*Y[3]*Y[2]*X[11]*X[7]+2*Y[13]*Y[15]*X[16]*X[8]-2*Y[7]*Y[11]*X[2]*X[3]-2*Y[8]*Y[16]*X[15]*X[13]+2*Y[1]*Y[4]*X[12]*X[9]+2*Y[5]*Y[6]*X[10]*X[14]);
  Z[12][10]:=1/(thc[7]*thc[8])*(-2*Y[10]*Y[12]*X[7]*X[8]-2*Y[15]*Y[2]*X[1]*X[5]-2*Y[16]*Y[11]*X[9]*X[14]-2*Y[6]*Y[4]*X[3]*X[13]+2*Y[14]*Y[9]*X[11]*X[16]+2*Y[13]*Y[3]*X[4]*X[6]+2*Y[8]*Y[7]*X[12]*X[10]+2*Y[5]*Y[1]*X[2]*X[15]);
  Z[12][11]:=1/(thc[3]*thc[1])*(-2*Y[11]*Y[12]*X[3]*X[1]+2*Y[16]*Y[10]*X[13]*X[5]-2*Y[4]*Y[2]*X[9]*X[7]+2*Y[6]*Y[15]*X[14]*X[8]+2*Y[7]*Y[9]*X[2]*X[4]-2*Y[8]*Y[14]*X[15]*X[6]+2*Y[1]*Y[3]*X[12]*X[11]-2*Y[5]*Y[13]*X[10]*X[16]);
  Z[12][12]:=1/(thc[1]^2)*(-Y[12]^2*X[1]^2-Y[10]^2*X[5]^2-Y[2]^2*X[7]^2-Y[15]^2*X[8]^2-Y[11]^2*X[3]^2-Y[16]^2*X[13]^2-Y[4]^2*X[9]^2-Y[6]^2*X[14]^2+Y[9]^2*X[4]^2+Y[14]^2*X[6]^2+Y[3]^2*X[11]^2+Y[13]^2*X[16]^2+Y[7]^2*X[2]^2+Y[8]^2*X[15]^2+Y[1]^2*X[12]^2+Y[5]^2*X[10]^2);
  Z[12][13]:=1/(thc[6]*thc[7])*(-2*Y[13]*Y[12]*X[6]*X[7]-2*Y[3]*Y[10]*X[4]*X[8]+2*Y[14]*Y[2]*X[16]*X[1]+2*Y[9]*Y[15]*X[11]*X[5]-2*Y[5]*Y[11]*X[15]*X[9]-2*Y[1]*Y[16]*X[2]*X[14]+2*Y[8]*Y[4]*X[10]*X[3]+2*Y[7]*Y[6]*X[12]*X[13]);
  Z[12][14]:=1/(thc[6]*thc[1])*(-2*Y[14]*Y[12]*X[6]*X[1]-2*Y[9]*Y[10]*X[4]*X[5]+2*Y[13]*Y[2]*X[16]*X[7]+2*Y[3]*Y[15]*X[11]*X[8]-2*Y[8]*Y[11]*X[15]*X[3]-2*Y[7]*Y[16]*X[2]*X[13]+2*Y[5]*Y[4]*X[10]*X[9]+2*Y[1]*Y[6]*X[12]*X[14]);
  Z[12][15]:=1/(thc[8]*thc[1])*(-2*Y[15]*Y[12]*X[8]*X[1]-2*Y[2]*Y[10]*X[7]*X[5]-2*Y[6]*Y[11]*X[14]*X[3]-2*Y[4]*Y[16]*X[9]*X[13]+2*Y[13]*Y[9]*X[16]*X[4]+2*Y[3]*Y[14]*X[11]*X[6]+2*Y[5]*Y[7]*X[10]*X[2]+2*Y[1]*Y[8]*X[12]*X[15]);
  Z[12][16]:=1/(thc[9]*thc[8])*(-2*Y[16]*Y[12]*X[9]*X[8]+2*Y[11]*Y[10]*X[14]*X[7]-2*Y[6]*Y[2]*X[3]*X[5]+2*Y[4]*Y[15]*X[13]*X[1]+2*Y[8]*Y[9]*X[12]*X[16]-2*Y[7]*Y[14]*X[10]*X[11]+2*Y[5]*Y[3]*X[2]*X[6]-2*Y[1]*Y[13]*X[15]*X[4]);

  Z[13][1]:=1/(thc[3]*thc[5])*(-2*Y[1]*Y[13]*X[3]*X[5]+2*Y[5]*Y[3]*X[13]*X[1]-2*Y[7]*Y[14]*X[9]*X[8]+2*Y[8]*Y[9]*X[14]*X[7]-2*Y[4]*Y[15]*X[2]*X[6]+2*Y[6]*Y[2]*X[15]*X[4]-2*Y[11]*Y[10]*X[12]*X[16]+2*Y[16]*Y[12]*X[10]*X[11]);
  Z[13][2]:=1/(thc[1]*thc[6])*(-2*Y[2]*Y[13]*X[1]*X[6]-2*Y[15]*Y[3]*X[5]*X[4]-2*Y[12]*Y[14]*X[7]*X[16]-2*Y[10]*Y[9]*X[8]*X[11]+2*Y[4]*Y[5]*X[3]*X[15]+2*Y[6]*Y[1]*X[13]*X[2]+2*Y[11]*Y[8]*X[9]*X[10]+2*Y[16]*Y[7]*X[14]*X[12]);
  Z[13][3]:=1/(thc[5]*thc[1])*(-2*Y[3]*Y[13]*X[5]*X[1]-2*Y[9]*Y[14]*X[8]*X[7]+2*Y[1]*Y[5]*X[13]*X[3]+2*Y[7]*Y[8]*X[14]*X[9]-2*Y[2]*Y[15]*X[6]*X[4]-2*Y[12]*Y[10]*X[16]*X[11]+2*Y[4]*Y[6]*X[15]*X[2]+2*Y[11]*Y[16]*X[10]*X[12]);
  Z[13][4]:=1/(thc[3]*thc[6])*(-2*Y[4]*Y[13]*X[3]*X[6]+2*Y[6]*Y[3]*X[13]*X[4]-2*Y[11]*Y[14]*X[9]*X[16]+2*Y[16]*Y[9]*X[14]*X[11]+2*Y[2]*Y[5]*X[1]*X[15]-2*Y[15]*Y[1]*X[5]*X[2]+2*Y[12]*Y[8]*X[7]*X[10]-2*Y[10]*Y[7]*X[8]*X[12]);
  Z[13][5]:=1/(thc[2]*thc[4])*(-2*Y[5]*Y[13]*X[2]*X[4]+2*Y[1]*Y[3]*X[15]*X[6]+2*Y[8]*Y[14]*X[12]*X[11]-2*Y[7]*Y[9]*X[10]*X[16]-2*Y[6]*Y[15]*X[3]*X[1]+2*Y[4]*Y[2]*X[13]*X[5]+2*Y[16]*Y[10]*X[9]*X[7]-2*Y[11]*Y[12]*X[14]*X[8]);
  Z[13][6]:=1/(thc[3]*thc[4])*(-2*Y[6]*Y[13]*X[3]*X[4]+2*Y[4]*Y[3]*X[13]*X[6]-2*Y[16]*Y[14]*X[9]*X[11]+2*Y[11]*Y[9]*X[14]*X[16]-2*Y[15]*Y[5]*X[1]*X[2]+2*Y[2]*Y[1]*X[5]*X[15]-2*Y[10]*Y[8]*X[7]*X[12]+2*Y[12]*Y[7]*X[8]*X[10]);
  Z[13][7]:=1/(thc[3]*thc[8])*(-2*Y[7]*Y[13]*X[3]*X[8]+2*Y[8]*Y[3]*X[13]*X[7]-2*Y[1]*Y[14]*X[9]*X[5]+2*Y[5]*Y[9]*X[14]*X[1]+2*Y[11]*Y[15]*X[2]*X[16]-2*Y[16]*Y[2]*X[15]*X[11]+2*Y[4]*Y[10]*X[12]*X[6]-2*Y[6]*Y[12]*X[10]*X[4]);
  Z[13][8]:=1/(thc[9]*thc[1])*(-2*Y[8]*Y[13]*X[9]*X[1]+2*Y[7]*Y[3]*X[14]*X[5]-2*Y[5]*Y[14]*X[3]*X[7]+2*Y[1]*Y[9]*X[13]*X[8]+2*Y[16]*Y[15]*X[12]*X[4]-2*Y[11]*Y[2]*X[10]*X[6]+2*Y[6]*Y[10]*X[2]*X[11]-2*Y[4]*Y[12]*X[15]*X[16]);
  Z[13][9]:=1/(thc[2]*thc[10])*(-2*Y[9]*Y[13]*X[2]*X[10]+2*Y[14]*Y[3]*X[15]*X[12]-2*Y[7]*Y[5]*X[4]*X[16]+2*Y[8]*Y[1]*X[6]*X[11]-2*Y[12]*Y[15]*X[3]*X[14]+2*Y[10]*Y[2]*X[13]*X[9]-2*Y[11]*Y[6]*X[1]*X[8]+2*Y[16]*Y[4]*X[5]*X[7]);
  Z[13][10]:=1/(thc[8]*thc[6])*(-2*Y[10]*Y[13]*X[8]*X[6]-2*Y[12]*Y[3]*X[7]*X[4]-2*Y[15]*Y[14]*X[5]*X[16]-2*Y[2]*Y[9]*X[1]*X[11]+2*Y[16]*Y[5]*X[14]*X[15]+2*Y[11]*Y[1]*X[9]*X[2]+2*Y[6]*Y[8]*X[13]*X[10]+2*Y[4]*Y[7]*X[3]*X[12]);
  Z[13][11]:=1/(thc[9]*thc[6])*(-2*Y[11]*Y[13]*X[9]*X[6]+2*Y[16]*Y[3]*X[14]*X[4]-2*Y[4]*Y[14]*X[3]*X[16]+2*Y[6]*Y[9]*X[13]*X[11]+2*Y[12]*Y[5]*X[7]*X[15]-2*Y[10]*Y[1]*X[8]*X[2]+2*Y[2]*Y[8]*X[1]*X[10]-2*Y[15]*Y[7]*X[5]*X[12]);
  Z[13][12]:=1/(thc[3]*thc[10])*(-2*Y[12]*Y[13]*X[3]*X[10]+2*Y[10]*Y[3]*X[13]*X[12]-2*Y[2]*Y[14]*X[9]*X[15]+2*Y[15]*Y[9]*X[14]*X[2]-2*Y[11]*Y[5]*X[1]*X[16]+2*Y[16]*Y[1]*X[5]*X[11]-2*Y[4]*Y[8]*X[7]*X[6]+2*Y[6]*Y[7]*X[8]*X[4]);
  Z[13][13]:=1/(thc[1]^2)*(-Y[13]^2*X[1]^2-Y[3]^2*X[5]^2-Y[14]^2*X[7]^2-Y[9]^2*X[8]^2+Y[5]^2*X[3]^2+Y[1]^2*X[13]^2+Y[8]^2*X[9]^2+Y[7]^2*X[14]^2-Y[15]^2*X[4]^2-Y[2]^2*X[6]^2-Y[10]^2*X[11]^2-Y[12]^2*X[16]^2+Y[6]^2*X[2]^2+Y[4]^2*X[15]^2+Y[16]^2*X[12]^2+Y[11]^2*X[10]^2);
  Z[13][14]:=1/(thc[5]*thc[8])*(-2*Y[14]*Y[13]*X[5]*X[8]-2*Y[9]*Y[3]*X[1]*X[7]+2*Y[8]*Y[5]*X[13]*X[14]+2*Y[7]*Y[1]*X[3]*X[9]+2*Y[10]*Y[15]*X[6]*X[16]+2*Y[12]*Y[2]*X[4]*X[11]-2*Y[16]*Y[6]*X[15]*X[10]-2*Y[11]*Y[4]*X[2]*X[12]);
  Z[13][15]:=1/(thc[1]*thc[4])*(-2*Y[15]*Y[13]*X[1]*X[4]-2*Y[2]*Y[3]*X[5]*X[6]-2*Y[10]*Y[14]*X[7]*X[11]-2*Y[12]*Y[9]*X[8]*X[16]+2*Y[6]*Y[5]*X[3]*X[2]+2*Y[4]*Y[1]*X[13]*X[15]+2*Y[16]*Y[8]*X[9]*X[12]+2*Y[11]*Y[7]*X[14]*X[10]);
  Z[13][16]:=1/(thc[5]*thc[10])*(-2*Y[16]*Y[13]*X[5]*X[10]+2*Y[11]*Y[3]*X[1]*X[12]-2*Y[6]*Y[14]*X[8]*X[15]+2*Y[4]*Y[9]*X[7]*X[2]+2*Y[10]*Y[5]*X[13]*X[16]-2*Y[12]*Y[1]*X[3]*X[11]+2*Y[15]*Y[8]*X[14]*X[6]-2*Y[2]*Y[7]*X[9]*X[4]);

  Z[14][1]:=1/(thc[9]*thc[5])*(-2*Y[1]*Y[14]*X[9]*X[5]+2*Y[5]*Y[9]*X[14]*X[1]-2*Y[7]*Y[13]*X[3]*X[8]+2*Y[8]*Y[3]*X[13]*X[7]-2*Y[4]*Y[10]*X[12]*X[6]+2*Y[6]*Y[12]*X[10]*X[4]-2*Y[11]*Y[15]*X[2]*X[16]+2*Y[16]*Y[2]*X[15]*X[11]);
  Z[14][2]:=1/(thc[7]*thc[6])*(-2*Y[2]*Y[14]*X[7]*X[6]-2*Y[15]*Y[9]*X[8]*X[4]-2*Y[12]*Y[13]*X[1]*X[16]-2*Y[10]*Y[3]*X[5]*X[11]+2*Y[4]*Y[8]*X[9]*X[15]+2*Y[6]*Y[7]*X[14]*X[2]+2*Y[11]*Y[5]*X[3]*X[10]+2*Y[16]*Y[1]*X[13]*X[12]);
  Z[14][3]:=1/(thc[7]*thc[5])*(-2*Y[3]*Y[14]*X[7]*X[5]-2*Y[13]*Y[9]*X[8]*X[1]+2*Y[1]*Y[8]*X[9]*X[13]+2*Y[5]*Y[7]*X[14]*X[3]-2*Y[2]*Y[10]*X[11]*X[6]-2*Y[15]*Y[12]*X[16]*X[4]+2*Y[4]*Y[16]*X[12]*X[15]+2*Y[6]*Y[11]*X[10]*X[2]);
  Z[14][4]:=1/(thc[9]*thc[6])*(-2*Y[4]*Y[14]*X[9]*X[6]+2*Y[6]*Y[9]*X[14]*X[4]-2*Y[11]*Y[13]*X[3]*X[16]+2*Y[16]*Y[3]*X[13]*X[11]+2*Y[2]*Y[8]*X[7]*X[15]-2*Y[15]*Y[7]*X[8]*X[2]+2*Y[12]*Y[5]*X[1]*X[10]-2*Y[10]*Y[1]*X[5]*X[12]);
  Z[14][5]:=1/(thc[3]*thc[7])*(-2*Y[5]*Y[14]*X[3]*X[7]+2*Y[1]*Y[9]*X[13]*X[8]-2*Y[8]*Y[13]*X[9]*X[1]+2*Y[7]*Y[3]*X[14]*X[5]-2*Y[6]*Y[10]*X[2]*X[11]+2*Y[4]*Y[12]*X[15]*X[16]-2*Y[16]*Y[15]*X[12]*X[4]+2*Y[11]*Y[2]*X[10]*X[6]);
  Z[14][6]:=1/(thc[9]*thc[4])*(-2*Y[6]*Y[14]*X[9]*X[4]+2*Y[4]*Y[9]*X[14]*X[6]-2*Y[16]*Y[13]*X[3]*X[11]+2*Y[11]*Y[3]*X[13]*X[16]-2*Y[15]*Y[8]*X[7]*X[2]+2*Y[2]*Y[7]*X[8]*X[15]-2*Y[10]*Y[5]*X[1]*X[12]+2*Y[12]*Y[1]*X[5]*X[10]);
  Z[14][7]:=1/(thc[9]*thc[8])*(-2*Y[7]*Y[14]*X[9]*X[8]+2*Y[8]*Y[9]*X[14]*X[7]-2*Y[1]*Y[13]*X[3]*X[5]+2*Y[5]*Y[3]*X[13]*X[1]+2*Y[11]*Y[10]*X[12]*X[16]-2*Y[16]*Y[12]*X[10]*X[11]+2*Y[4]*Y[15]*X[2]*X[6]-2*Y[6]*Y[2]*X[15]*X[4]);
  Z[14][8]:=1/(thc[2]*thc[4])*(-2*Y[8]*Y[14]*X[2]*X[4]+2*Y[7]*Y[9]*X[15]*X[6]+2*Y[5]*Y[13]*X[12]*X[11]-2*Y[1]*Y[3]*X[10]*X[16]+2*Y[16]*Y[10]*X[3]*X[1]-2*Y[11]*Y[12]*X[13]*X[5]-2*Y[6]*Y[15]*X[9]*X[7]+2*Y[4]*Y[2]*X[14]*X[8]);
  Z[14][9]:=1/(thc[6]*thc[4])*(-2*Y[9]*Y[14]*X[6]*X[4]+2*Y[3]*Y[13]*X[16]*X[11]+2*Y[7]*Y[8]*X[15]*X[2]-2*Y[1]*Y[5]*X[10]*X[12]+2*Y[12]*Y[10]*X[5]*X[1]-2*Y[2]*Y[15]*X[8]*X[7]-2*Y[11]*Y[16]*X[13]*X[3]+2*Y[4]*Y[6]*X[14]*X[9]);
  Z[14][10]:=1/(thc[5]*thc[6])*(-2*Y[10]*Y[14]*X[5]*X[6]-2*Y[12]*Y[9]*X[1]*X[4]-2*Y[15]*Y[13]*X[8]*X[16]-2*Y[2]*Y[3]*X[7]*X[11]+2*Y[16]*Y[8]*X[13]*X[15]+2*Y[11]*Y[7]*X[3]*X[2]+2*Y[6]*Y[5]*X[14]*X[10]+2*Y[4]*Y[1]*X[9]*X[12]);
  Z[14][11]:=1/(thc[7]*thc[10])*(-2*Y[11]*Y[14]*X[7]*X[10]+2*Y[16]*Y[9]*X[8]*X[12]-2*Y[4]*Y[13]*X[1]*X[15]+2*Y[6]*Y[3]*X[5]*X[2]-2*Y[12]*Y[8]*X[9]*X[16]+2*Y[10]*Y[7]*X[14]*X[11]-2*Y[2]*Y[5]*X[3]*X[6]+2*Y[15]*Y[1]*X[13]*X[4]);
  Z[14][12]:=1/(thc[1]*thc[6])*(-2*Y[12]*Y[14]*X[1]*X[6]-2*Y[10]*Y[9]*X[5]*X[4]-2*Y[2]*Y[13]*X[7]*X[16]-2*Y[15]*Y[3]*X[8]*X[11]+2*Y[11]*Y[8]*X[3]*X[15]+2*Y[16]*Y[7]*X[13]*X[2]+2*Y[4]*Y[5]*X[9]*X[10]+2*Y[6]*Y[1]*X[14]*X[12]);
  Z[14][13]:=1/(thc[9]*thc[3])*(-2*Y[13]*Y[14]*X[9]*X[3]+2*Y[3]*Y[9]*X[14]*X[13]+2*Y[5]*Y[8]*X[7]*X[1]-2*Y[1]*Y[7]*X[8]*X[5]-2*Y[15]*Y[10]*X[12]*X[2]+2*Y[2]*Y[12]*X[10]*X[15]+2*Y[6]*Y[16]*X[11]*X[4]-2*Y[4]*Y[11]*X[16]*X[6]);
  Z[14][14]:=1/(thc[1]^2)*(-Y[14]^2*X[1]^2-Y[9]^2*X[5]^2-Y[13]^2*X[7]^2-Y[3]^2*X[8]^2+Y[8]^2*X[3]^2+Y[7]^2*X[13]^2+Y[5]^2*X[9]^2+Y[1]^2*X[14]^2+Y[10]^2*X[4]^2+Y[12]^2*X[6]^2+Y[15]^2*X[11]^2+Y[2]^2*X[16]^2-Y[16]^2*X[2]^2-Y[11]^2*X[15]^2-Y[6]^2*X[12]^2-Y[4]^2*X[10]^2);
  Z[14][15]:=1/(thc[8]*thc[6])*(-2*Y[15]*Y[14]*X[8]*X[6]-2*Y[2]*Y[9]*X[7]*X[4]-2*Y[10]*Y[13]*X[5]*X[16]-2*Y[12]*Y[3]*X[1]*X[11]+2*Y[6]*Y[8]*X[14]*X[15]+2*Y[4]*Y[7]*X[9]*X[2]+2*Y[16]*Y[5]*X[13]*X[10]+2*Y[11]*Y[1]*X[3]*X[12]);
  Z[14][16]:=1/(thc[1]*thc[2])*(-2*Y[16]*Y[14]*X[1]*X[2]+2*Y[11]*Y[9]*X[5]*X[15]-2*Y[6]*Y[13]*X[7]*X[12]+2*Y[4]*Y[3]*X[8]*X[10]-2*Y[10]*Y[8]*X[3]*X[4]+2*Y[12]*Y[7]*X[13]*X[6]-2*Y[15]*Y[5]*X[9]*X[11]+2*Y[2]*Y[1]*X[14]*X[16]);

  Z[15][1]:=1/(thc[10]*thc[7])*(-2*Y[1]*Y[15]*X[10]*X[7]+2*Y[5]*Y[2]*X[12]*X[8]+2*Y[7]*Y[10]*X[15]*X[1]-2*Y[8]*Y[12]*X[2]*X[5]-2*Y[3]*Y[6]*X[16]*X[9]+2*Y[13]*Y[4]*X[11]*X[14]+2*Y[9]*Y[16]*X[6]*X[3]-2*Y[14]*Y[11]*X[4]*X[13]);
  Z[15][2]:=1/(thc[5]*thc[1])*(-2*Y[2]*Y[15]*X[5]*X[1]-2*Y[12]*Y[10]*X[8]*X[7]+2*Y[4]*Y[6]*X[13]*X[3]+2*Y[11]*Y[16]*X[14]*X[9]-2*Y[3]*Y[13]*X[6]*X[4]-2*Y[9]*Y[14]*X[16]*X[11]+2*Y[1]*Y[5]*X[15]*X[2]+2*Y[7]*Y[8]*X[10]*X[12]);
  Z[15][3]:=1/(thc[4]*thc[5])*(-2*Y[3]*Y[15]*X[4]*X[5]-2*Y[13]*Y[2]*X[6]*X[1]+2*Y[9]*Y[10]*X[11]*X[8]+2*Y[14]*Y[12]*X[16]*X[7]+2*Y[1]*Y[6]*X[2]*X[13]+2*Y[5]*Y[4]*X[15]*X[3]-2*Y[7]*Y[16]*X[12]*X[14]-2*Y[8]*Y[11]*X[10]*X[9]);
  Z[15][4]:=1/(thc[3]*thc[5])*(-2*Y[4]*Y[15]*X[3]*X[5]+2*Y[6]*Y[2]*X[13]*X[1]-2*Y[11]*Y[10]*X[9]*X[8]+2*Y[16]*Y[12]*X[14]*X[7]-2*Y[1]*Y[13]*X[2]*X[6]+2*Y[5]*Y[3]*X[15]*X[4]-2*Y[7]*Y[14]*X[12]*X[16]+2*Y[8]*Y[9]*X[10]*X[11]);
  Z[15][5]:=1/(thc[4]*thc[3])*(-2*Y[5]*Y[15]*X[4]*X[3]+2*Y[1]*Y[2]*X[6]*X[13]+2*Y[8]*Y[10]*X[11]*X[9]-2*Y[7]*Y[12]*X[16]*X[14]-2*Y[13]*Y[6]*X[2]*X[1]+2*Y[3]*Y[4]*X[15]*X[5]+2*Y[14]*Y[16]*X[12]*X[7]-2*Y[9]*Y[11]*X[10]*X[8]);
  Z[15][6]:=1/(thc[2]*thc[4])*(-2*Y[6]*Y[15]*X[2]*X[4]+2*Y[4]*Y[2]*X[15]*X[6]+2*Y[16]*Y[10]*X[12]*X[11]-2*Y[11]*Y[12]*X[10]*X[16]-2*Y[5]*Y[13]*X[3]*X[1]+2*Y[1]*Y[3]*X[13]*X[5]+2*Y[8]*Y[14]*X[9]*X[7]-2*Y[7]*Y[9]*X[14]*X[8]);
  Z[15][7]:=1/(thc[6]*thc[9])*(-2*Y[7]*Y[15]*X[6]*X[9]+2*Y[8]*Y[2]*X[4]*X[14]+2*Y[1]*Y[10]*X[16]*X[3]-2*Y[5]*Y[12]*X[11]*X[13]+2*Y[9]*Y[6]*X[15]*X[7]-2*Y[14]*Y[4]*X[2]*X[8]-2*Y[3]*Y[16]*X[10]*X[1]+2*Y[13]*Y[11]*X[12]*X[5]);
  Z[15][8]:=1/(thc[4]*thc[9])*(-2*Y[8]*Y[15]*X[4]*X[9]+2*Y[7]*Y[2]*X[6]*X[14]+2*Y[5]*Y[10]*X[11]*X[3]-2*Y[1]*Y[12]*X[16]*X[13]-2*Y[14]*Y[6]*X[2]*X[7]+2*Y[9]*Y[4]*X[15]*X[8]+2*Y[13]*Y[16]*X[12]*X[1]-2*Y[3]*Y[11]*X[10]*X[5]);
  Z[15][9]:=1/(thc[6]*thc[7])*(-2*Y[9]*Y[15]*X[6]*X[7]-2*Y[14]*Y[2]*X[4]*X[8]+2*Y[3]*Y[10]*X[16]*X[1]+2*Y[13]*Y[12]*X[11]*X[5]+2*Y[7]*Y[6]*X[15]*X[9]+2*Y[8]*Y[4]*X[2]*X[14]-2*Y[1]*Y[16]*X[10]*X[3]-2*Y[5]*Y[11]*X[12]*X[13]);
  Z[15][10]:=1/(thc[1]*thc[7])*(-2*Y[10]*Y[15]*X[1]*X[7]-2*Y[12]*Y[2]*X[5]*X[8]+2*Y[16]*Y[6]*X[3]*X[9]+2*Y[11]*Y[4]*X[13]*X[14]-2*Y[14]*Y[13]*X[4]*X[11]-2*Y[9]*Y[3]*X[6]*X[16]+2*Y[8]*Y[5]*X[2]*X[12]+2*Y[7]*Y[1]*X[15]*X[10]);
  Z[15][11]:=1/(thc[9]*thc[5])*(-2*Y[11]*Y[15]*X[9]*X[5]+2*Y[16]*Y[2]*X[14]*X[1]-2*Y[4]*Y[10]*X[3]*X[8]+2*Y[6]*Y[12]*X[13]*X[7]-2*Y[7]*Y[13]*X[12]*X[6]+2*Y[8]*Y[3]*X[10]*X[4]-2*Y[1]*Y[14]*X[2]*X[16]+2*Y[5]*Y[9]*X[15]*X[11]);
  Z[15][12]:=1/(thc[1]*thc[8])*(-2*Y[12]*Y[15]*X[1]*X[8]-2*Y[10]*Y[2]*X[5]*X[7]+2*Y[11]*Y[6]*X[3]*X[14]+2*Y[16]*Y[4]*X[13]*X[9]-2*Y[9]*Y[13]*X[4]*X[16]-2*Y[14]*Y[3]*X[6]*X[11]+2*Y[7]*Y[5]*X[2]*X[10]+2*Y[8]*Y[1]*X[15]*X[12]);
  Z[15][13]:=1/(thc[2]*thc[3])*(-2*Y[13]*Y[15]*X[2]*X[3]+2*Y[3]*Y[2]*X[15]*X[13]+2*Y[14]*Y[10]*X[12]*X[9]-2*Y[9]*Y[12]*X[10]*X[14]+2*Y[5]*Y[6]*X[4]*X[1]-2*Y[1]*Y[4]*X[6]*X[5]-2*Y[8]*Y[16]*X[11]*X[7]+2*Y[7]*Y[11]*X[16]*X[8]);
  Z[15][14]:=1/(thc[6]*thc[8])*(-2*Y[14]*Y[15]*X[6]*X[8]-2*Y[9]*Y[2]*X[4]*X[7]+2*Y[13]*Y[10]*X[16]*X[5]+2*Y[3]*Y[12]*X[11]*X[1]+2*Y[8]*Y[6]*X[15]*X[14]+2*Y[7]*Y[4]*X[2]*X[9]-2*Y[5]*Y[16]*X[10]*X[13]-2*Y[1]*Y[11]*X[12]*X[3]);
  Z[15][15]:=1/(thc[1]^2)*(-Y[15]^2*X[1]^2-Y[2]^2*X[5]^2-Y[10]^2*X[7]^2-Y[12]^2*X[8]^2+Y[6]^2*X[3]^2+Y[4]^2*X[13]^2+Y[16]^2*X[9]^2+Y[11]^2*X[14]^2-Y[13]^2*X[4]^2-Y[3]^2*X[6]^2-Y[14]^2*X[11]^2-Y[9]^2*X[16]^2+Y[5]^2*X[2]^2+Y[1]^2*X[15]^2+Y[8]^2*X[12]^2+Y[7]^2*X[10]^2);
  Z[15][16]:=1/(thc[3]*thc[7])*(-2*Y[16]*Y[15]*X[3]*X[7]+2*Y[11]*Y[2]*X[13]*X[8]-2*Y[6]*Y[10]*X[9]*X[1]+2*Y[4]*Y[12]*X[14]*X[5]-2*Y[8]*Y[13]*X[2]*X[11]+2*Y[7]*Y[3]*X[15]*X[16]-2*Y[5]*Y[14]*X[12]*X[4]+2*Y[1]*Y[9]*X[10]*X[6]);

  Z[16][1]:=1/(thc[10]*thc[3])*(-2*Y[1]*Y[16]*X[10]*X[3]+2*Y[5]*Y[11]*X[12]*X[13]+2*Y[7]*Y[6]*X[15]*X[9]-2*Y[8]*Y[4]*X[2]*X[14]+2*Y[3]*Y[10]*X[16]*X[1]-2*Y[13]*Y[12]*X[11]*X[5]-2*Y[9]*Y[15]*X[6]*X[7]+2*Y[14]*Y[2]*X[4]*X[8]);
  Z[16][2]:=1/(thc[4]*thc[10])*(-2*Y[2]*Y[16]*X[4]*X[10]-2*Y[15]*Y[11]*X[6]*X[12]+2*Y[12]*Y[6]*X[11]*X[15]+2*Y[10]*Y[4]*X[16]*X[2]-2*Y[3]*Y[8]*X[1]*X[14]-2*Y[13]*Y[7]*X[5]*X[9]+2*Y[9]*Y[5]*X[7]*X[13]+2*Y[14]*Y[1]*X[8]*X[3]);
  Z[16][3]:=1/(thc[6]*thc[9])*(-2*Y[3]*Y[16]*X[6]*X[9]-2*Y[13]*Y[11]*X[4]*X[14]+2*Y[9]*Y[6]*X[16]*X[3]+2*Y[14]*Y[4]*X[11]*X[13]-2*Y[1]*Y[10]*X[15]*X[7]-2*Y[5]*Y[12]*X[2]*X[8]+2*Y[7]*Y[15]*X[10]*X[1]+2*Y[8]*Y[2]*X[12]*X[5]);
  Z[16][4]:=1/(thc[7]*thc[5])*(-2*Y[4]*Y[16]*X[7]*X[5]-2*Y[6]*Y[11]*X[8]*X[1]-2*Y[2]*Y[10]*X[9]*X[13]-2*Y[15]*Y[12]*X[14]*X[3]+2*Y[1]*Y[8]*X[11]*X[6]+2*Y[5]*Y[7]*X[16]*X[4]+2*Y[3]*Y[14]*X[12]*X[15]+2*Y[13]*Y[9]*X[10]*X[2]);
  Z[16][5]:=1/(thc[6]*thc[8])*(-2*Y[5]*Y[16]*X[6]*X[8]-2*Y[1]*Y[11]*X[4]*X[7]+2*Y[8]*Y[6]*X[16]*X[5]+2*Y[7]*Y[4]*X[11]*X[1]+2*Y[13]*Y[10]*X[15]*X[14]+2*Y[3]*Y[12]*X[2]*X[9]-2*Y[14]*Y[15]*X[10]*X[13]-2*Y[9]*Y[2]*X[12]*X[3]);
  Z[16][6]:=1/(thc[9]*thc[3])*(-2*Y[6]*Y[16]*X[9]*X[3]+2*Y[4]*Y[11]*X[14]*X[13]+2*Y[15]*Y[10]*X[7]*X[1]-2*Y[2]*Y[12]*X[8]*X[5]+2*Y[5]*Y[8]*X[12]*X[2]-2*Y[1]*Y[7]*X[10]*X[15]-2*Y[13]*Y[14]*X[11]*X[4]+2*Y[3]*Y[9]*X[16]*X[6]);
  Z[16][7]:=1/(thc[10]*thc[9])*(-2*Y[7]*Y[16]*X[10]*X[9]+2*Y[8]*Y[11]*X[12]*X[14]+2*Y[1]*Y[6]*X[15]*X[3]-2*Y[5]*Y[4]*X[2]*X[13]+2*Y[9]*Y[10]*X[16]*X[7]-2*Y[14]*Y[12]*X[11]*X[8]-2*Y[3]*Y[15]*X[6]*X[1]+2*Y[13]*Y[2]*X[4]*X[5]);
  Z[16][8]:=1/(thc[6]*thc[5])*(-2*Y[8]*Y[16]*X[6]*X[5]-2*Y[7]*Y[11]*X[4]*X[1]+2*Y[5]*Y[6]*X[16]*X[8]+2*Y[1]*Y[4]*X[11]*X[7]+2*Y[14]*Y[10]*X[15]*X[13]+2*Y[9]*Y[12]*X[2]*X[3]-2*Y[13]*Y[15]*X[10]*X[14]-2*Y[3]*Y[2]*X[12]*X[9]);
  Z[16][9]:=1/(thc[2]*thc[5])*(-2*Y[9]*Y[16]*X[2]*X[5]-2*Y[14]*Y[11]*X[15]*X[1]+2*Y[3]*Y[6]*X[12]*X[8]+2*Y[13]*Y[4]*X[10]*X[7]-2*Y[7]*Y[10]*X[4]*X[13]-2*Y[8]*Y[12]*X[6]*X[3]+2*Y[1]*Y[15]*X[11]*X[14]+2*Y[5]*Y[2]*X[16]*X[9]);
  Z[16][10]:=1/(thc[1]*thc[3])*(-2*Y[10]*Y[16]*X[1]*X[3]-2*Y[12]*Y[11]*X[5]*X[13]-2*Y[15]*Y[6]*X[7]*X[9]-2*Y[2]*Y[4]*X[8]*X[14]+2*Y[14]*Y[8]*X[4]*X[2]+2*Y[9]*Y[7]*X[6]*X[15]+2*Y[13]*Y[5]*X[11]*X[12]+2*Y[3]*Y[1]*X[16]*X[10]);
  Z[16][11]:=1/(thc[5]*thc[1])*(-2*Y[11]*Y[16]*X[5]*X[1]-2*Y[4]*Y[6]*X[8]*X[7]-2*Y[12]*Y[10]*X[13]*X[3]-2*Y[2]*Y[15]*X[14]*X[9]+2*Y[7]*Y[8]*X[6]*X[4]+2*Y[1]*Y[5]*X[16]*X[11]+2*Y[9]*Y[14]*X[15]*X[2]+2*Y[3]*Y[13]*X[10]*X[12]);
  Z[16][12]:=1/(thc[8]*thc[9])*(-2*Y[12]*Y[16]*X[8]*X[9]-2*Y[10]*Y[11]*X[7]*X[14]-2*Y[2]*Y[6]*X[5]*X[3]-2*Y[15]*Y[4]*X[1]*X[13]+2*Y[9]*Y[8]*X[16]*X[12]+2*Y[14]*Y[7]*X[11]*X[10]+2*Y[3]*Y[5]*X[6]*X[2]+2*Y[13]*Y[1]*X[4]*X[15]);
  Z[16][13]:=1/(thc[2]*thc[7])*(-2*Y[13]*Y[16]*X[2]*X[7]-2*Y[3]*Y[11]*X[15]*X[8]+2*Y[14]*Y[6]*X[12]*X[1]+2*Y[9]*Y[4]*X[10]*X[5]-2*Y[5]*Y[10]*X[4]*X[9]-2*Y[1]*Y[12]*X[6]*X[14]+2*Y[8]*Y[15]*X[11]*X[3]+2*Y[7]*Y[2]*X[16]*X[13]);
  Z[16][14]:=1/(thc[10]*thc[8])*(-2*Y[14]*Y[16]*X[10]*X[8]-2*Y[9]*Y[11]*X[12]*X[7]+2*Y[13]*Y[6]*X[15]*X[5]+2*Y[3]*Y[4]*X[2]*X[1]+2*Y[8]*Y[10]*X[16]*X[14]+2*Y[7]*Y[12]*X[11]*X[9]-2*Y[5]*Y[15]*X[6]*X[13]-2*Y[1]*Y[2]*X[4]*X[3]);
  Z[16][15]:=1/(thc[1]*thc[9])*(-2*Y[15]*Y[16]*X[1]*X[9]-2*Y[2]*Y[11]*X[5]*X[14]-2*Y[10]*Y[6]*X[7]*X[3]-2*Y[12]*Y[4]*X[8]*X[13]+2*Y[13]*Y[8]*X[4]*X[12]+2*Y[3]*Y[7]*X[6]*X[10]+2*Y[14]*Y[5]*X[11]*X[2]+2*Y[9]*Y[1]*X[16]*X[15]);
  Z[16][16]:=1/(thc[1]^2)*(-Y[16]^2*X[1]^2-Y[11]^2*X[5]^2-Y[6]^2*X[7]^2-Y[4]^2*X[8]^2+Y[10]^2*X[3]^2+Y[12]^2*X[13]^2+Y[15]^2*X[9]^2+Y[2]^2*X[14]^2+Y[8]^2*X[4]^2+Y[7]^2*X[6]^2+Y[5]^2*X[11]^2+Y[1]^2*X[16]^2-Y[14]^2*X[2]^2-Y[9]^2*X[15]^2-Y[13]^2*X[12]^2-Y[3]^2*X[10]^2);

  for i in [1..16] do for j in [1..16] do Z[i][j]/:=4; end for; end for;

  return Z;
end function;



/*
Let thc be the theta constants numbered as in Gaudry
Let D1,D2 two divisors on the jacobian
Let Ros be the Rosenhain invariants of the curve. We assume that Ros[2]-Ros[1]=1

First compute the images X,Y of D1 and D2 in theta coordinates.

return the 16 Baily's additions law of X and Y as a list of list Z
Z[k] is th k-th law (wich corresponds to the projective factor
        theta_k(2(x-y),Om)=0
Z[k][l] is the l-th coordinates of X+Y given by the k-th law
 */
function BailyAdd_fromMumford(D1,D2,thc,Ros,theta_const_group)
  X:=MumfordToTheta(D1,theta_const_group,Ros);
  Y:=MumfordToTheta(D2,theta_const_group,Ros);
  return BailyAdd_fromTheta(X,Y,thc),X,Y;
end function;


/*
Let thc be the theta constants numbered as in Gaudry
Let X,Y be theta points in Gaudry numbering

Let lambda be coefficients to the bailly additions law P_k
return the addition law sum_k lambda[k] P_k
*/
function additionP_fromTheta(X,Y,thc,gamma)
  Z:=BailyAdd_fromTheta(X,Y,thc);
  return [ &+[gamma[k]*Z[k][l] : k in [1..16]] : l in [1..16]];
end function;

/*
Let thc be the theta constants numbered as in Gaudry
Let D1,D2 two divisors on the jacobian
Let Ros be the Rosenhain invariants of the curve. We assume that Ros[2]-Ros[1]=1

Let lambda be coefficients to the bailly additions law P_k
return the addition law sum_k lambda[k] P_k
 */
intrinsic additionP_fromMumford(D1 :: JacHypPt,D2 :: JacHypPt,thc :: SeqEnum ,Ros ::SeqEnum,theta_const_group :: Rec,gamma :: SeqEnum) -> SeqEnum
  {Some description goes here.}
  Z:=BailyAdd_fromMumford(D1,D2,thc,Ros,theta_const_group);
  return [ &+[gamma[k]*Z[k][l] : k in [1..16]] : l in [1..16]];
end intrinsic;


/*
Let JacC be the jacobian of an hyperelliptic curve over k
Find the alpha_i needed for the complete addition law.
*/
intrinsic FindAlpha(JacC :: JacHyp,k:: FldFin) -> SeqEnum
  {Some description goes here.}

  f:=HyperellipticPolynomials(Curve(JacC));

  q:=#k;
  K:=ext<k|2>;
  L:=ext<K|2>;

  repeat
    x0:=Random(K);
  until not IsSquare(K!Evaluate(f,x0)) and not IsCoercible(k,x0);

  y0:=Sqrt(L!Evaluate(f,x0));

  CL:=BaseChange(Curve(JacC),L);
  JacCL:=BaseChange(JacC,L);
  PLinf:=PointsAtInfinity(CL)[1];

  P:=CL![L!x0,y0];

  Pbar:=CL![L!x0,-y0];
  Psigma:=CL![L!x0^q,y0^q];
  Pbarsigma:=CL![L!x0^q,-y0^q];

  al0:=JacCL![[P,Psigma],[PLinf,PLinf]];
  al1:=JacCL![[Psigma,Pbar],[PLinf,PLinf]];
  al2:=JacCL![[Pbar,Pbarsigma],[PLinf,PLinf]];
  al3:=JacCL![[Pbarsigma,P],[PLinf,PLinf]];

  return [al0,al1,al2,al3];
end intrinsic;


/*
Let JacC be a jacobian of an hyperelliptic curve over a field k with analytic theta constants thc
Let alpha a list of elements of JacCL which is JacC considered over an extension field L of k.

Let P be an addition law that can be writen as P=sum_k lambda[k]*P_k where lambda[k] is in k and such that P is not defined over the divisor D0 = sum_k Theta_{alpha[k]}.

Find the coefficients lambda. We send random points of D0 to get a system in lambda and then we solve it.
 */
intrinsic CoefficientLawP(JacC :: JacHyp,Ros :: SeqEnum,thc :: Rec,alpha :: SeqEnum) -> SeqEnum
  {Some description goes here.}

  k:=BaseField(JacC);
  C:=Curve(JacC);

  JacCL:=Parent(alpha[1]);
  CL:=Curve(JacCL);
  PLinf:=PointsAtInfinity(CL)[1];
  L:=BaseField(JacCL);

  Mat    :=ZeroMatrix(L,1,10);
//  Mat_odd:=ZeroMatrix(L,1,6 );

  for i in [1..10] do
    print "Find line",i,"of the linear system";
    P:=Random(C);
    a:=alpha[Random([1..4])];
    D:=JacCL![[CL!P],[PLinf]] + a;
    t:=MumfordToTheta(D,thc,Ros);
    Mat    :=VerticalJoin(Mat    ,Matrix(L,1,10,t[1 ..10]));
//    Mat_odd:=VerticalJoin(Mat_odd,Matrix(L,1, 6,t[11..16]));
  end for;

  r:=Rank(Mat);
  print "The rank of the system is",r;

  while r lt 9 do
    P:=Random(C);
    a:=alpha[Random([1..4])];
    D:=JacCL![[CL!P],[PLinf]] + a;
    t:=MumfordToTheta(D,thc,Ros);
    Mat    :=VerticalJoin(Mat    ,Matrix(L,1,10,t[1 ..10]));
//   Mat_odd:=VerticalJoin(Mat_odd,Matrix(L,1, 6,t[11..16]));
    r:=Rank(Mat);
    print "The rank of the system is",r;
  end while;


//  assert Rank(Mat_odd) eq 6;


  print "Solving the system";
  E,T:=EchelonForm(Mat);
  Ec:=Submatrix(E,1,1,10,10);
  assert Submatrix(E,10,1,Nrows(E)-10,10) eq ZeroMatrix(L,Nrows(E)-10,10);
  assert IsConsistent(Ec,ZeroMatrix(L,1,10));

  l,ker:=Solution(Transpose(Ec), ZeroMatrix(L,1,10));
  assert Eltseq(l) eq [L!0 : x in [1..10]];
  v:=Eltseq(Basis(ker)[1]); vk:=[k!a : a in v];

  return vk cat [k!0 : x in [1..6]];
end intrinsic;

